#!/usr/bin/python

# Copyright (c) 2015, Sebastian Kornehl <sebastian.kornehl@asideas.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: datadog_monitor
short_description: Manages Datadog monitors
description:
  - Manages monitors within Datadog.
  - Options as described on https://docs.datadoghq.com/api/.
author: Sebastian Kornehl (@skornehl)
requirements: [datadog]
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  api_key:
    description:
      - Your Datadog API key.
    required: true
    type: str
  api_host:
    description:
      - The URL to the Datadog API. Default value is V(https://api.datadoghq.com).
      - This value can also be set with the E(DATADOG_HOST) environment variable.
    type: str
    version_added: '0.2.0'
  app_key:
    description:
      - Your Datadog app key.
    required: true
    type: str
  state:
    description:
      - The designated state of the monitor.
      - The values V(mute) and V(unmute) are deprecated since community.general 13.0.0 and will be removed in 15.0.0.
        Use the M(community.general.datadog_downtime) module to manage monitor downtimes instead.
    required: true
    choices: ['present', 'absent', 'mute', 'unmute']
    type: str
  tags:
    description:
      - A list of tags to associate with your monitor when creating or updating.
      - This can help you categorize and filter monitors.
    type: list
    elements: str
  type:
    description:
      - The type of the monitor.
      - The types V(query alert), V(trace-analytics alert) and V(rum alert) were added in community.general 2.1.0.
      - The type V(composite) was added in community.general 3.4.0.
      - The type V(event-v2 alert) was added in community.general 4.8.0.
    choices:
      - metric alert
      - service check
      - event alert
      - event-v2 alert
      - process alert
      - log alert
      - query alert
      - trace-analytics alert
      - rum alert
      - composite
    type: str
  query:
    description:
      - The monitor query to notify on.
      - Syntax varies depending on what type of monitor you are creating.
    type: str
  name:
    description:
      - The name of the alert.
    required: true
    type: str
  notification_message:
    description:
      - A message to include with notifications for this monitor.
      - Email notifications can be sent to specific users by using the same '@username' notation as events.
      - Monitor message template variables can be accessed by using double square brackets, in other words C([[) and C(]]).
    type: str
  silenced:
    type: dict
    description:
      - Dictionary of scopes to silence, with timestamps or None.
      - Each scope is muted until the given POSIX timestamp or forever if the value is V(None).
  notify_no_data:
    description:
      - Whether this monitor notifies when data stops reporting.
    type: bool
    default: false
  no_data_timeframe:
    description:
      - The number of minutes before a monitor notifies when data stops reporting.
      - Must be at least 2x the monitor timeframe for metric alerts or 2 minutes for service checks.
      - If not specified, it defaults to 2x timeframe for metric, 2 minutes for service.
    type: str
  timeout_h:
    description:
      - The number of hours of the monitor not reporting data before it automatically resolves from a triggered state.
    type: str
  renotify_interval:
    description:
      - The number of minutes after the last notification before a monitor re-notifies on the current status.
      - It only re-notifies if it is not resolved.
    type: str
  escalation_message:
    description:
      - A message to include with a re-notification. Supports the '@username' notification we allow elsewhere.
      - Not applicable if O(renotify_interval=none).
    type: str
  notify_audit:
    description:
      - Whether tagged users are notified on changes to this monitor.
    type: bool
    default: false
  thresholds:
    type: dict
    description:
      - A dictionary of thresholds by status.
      - Only available for service checks and metric alerts.
      - Because each of them can have multiple thresholds, we do not define them directly in the query.
      - "If not specified, it defaults to: V({'ok': 1, 'critical': 1, 'warning': 1})."
  locked:
    description:
      - Whether changes to this monitor should be restricted to the creator or admins.
    type: bool
    default: false
  require_full_window:
    description:
      - Whether this monitor needs a full window of data before it gets evaluated.
      - We highly recommend you set this to V(false) for sparse metrics, otherwise some evaluations are skipped.
    type: bool
  new_host_delay:
    description:
      - A positive integer representing the number of seconds to wait before evaluating the monitor for new hosts.
      - This gives the host time to fully initialize.
    type: str
  evaluation_delay:
    description:
      - Time to delay evaluation (in seconds).
      - Effective for sparse values.
    type: str
  id:
    description:
      - The ID of the alert.
      - If set, it is used instead of O(name) to locate the alert.
    type: str
  include_tags:
    description:
      - Whether notifications from this monitor automatically inserts its triggering tags into the title.
    type: bool
    default: true
    version_added: 1.3.0
  priority:
    description:
      - Integer from V(1) (high) to V(5) (low) indicating alert severity.
    type: int
    version_added: 4.6.0
  notification_preset_name:
    description:
      - Toggles the display of additional content sent in the monitor notification.
    choices:
      - show_all
      - hide_query
      - hide_handles
      - hide_all
    type: str
    version_added: 7.1.0
  renotify_occurrences:
    description:
      - The number of times re-notification messages should be sent on the current status at the provided re-notification
        interval.
    type: int
    version_added: 7.1.0
  renotify_statuses:
    description:
      - The types of monitor statuses for which re-notification messages are sent.
    choices:
      - alert
      - warn
      - no data
    type: list
    elements: str
    version_added: 7.1.0
"""

EXAMPLES = r"""
- name: Create a metric monitor
  community.general.datadog_monitor:
    type: "metric alert"
    name: "Test monitor"
    state: "present"
    renotify_interval: 30
    renotify_occurrences: 1
    renotify_statuses: ["warn"]
    notification_preset_name: "show_all"
    query: "datadog.agent.up.over('host:host1').last(2).count_by_status()"
    notification_message: "Host [[host.name]] with IP [[host.ip]] is failing to report to datadog."
    api_key: "9775a026f1ca7d1c6c5af9d94d9595a4"
    app_key: "87ce4a24b5553d2e482ea8a8500e71b8ad4554ff"

- name: Deletes a monitor
  community.general.datadog_monitor:
    name: "Test monitor"
    state: "absent"
    api_key: "9775a026f1ca7d1c6c5af9d94d9595a4"
    app_key: "87ce4a24b5553d2e482ea8a8500e71b8ad4554ff"

- name: Mutes a monitor (deprecated, use community.general.datadog_downtime instead)
  community.general.datadog_monitor:
    name: "Test monitor"
    state: "mute"
    silenced: '{"*":None}'
    api_key: "9775a026f1ca7d1c6c5af9d94d9595a4"
    app_key: "87ce4a24b5553d2e482ea8a8500e71b8ad4554ff"

- name: Unmutes a monitor (deprecated, use community.general.datadog_downtime instead)
  community.general.datadog_monitor:
    name: "Test monitor"
    state: "unmute"
    api_key: "9775a026f1ca7d1c6c5af9d94d9595a4"
    app_key: "87ce4a24b5553d2e482ea8a8500e71b8ad4554ff"

- name: Use datadoghq.eu platform instead of datadoghq.com
  community.general.datadog_monitor:
    name: "Test monitor"
    state: "absent"
    api_host: https://api.datadoghq.eu
    api_key: "9775a026f1ca7d1c6c5af9d94d9595a4"
    app_key: "87ce4a24b5553d2e482ea8a8500e71b8ad4554ff"
"""

import traceback

# Import Datadog
DATADOG_IMP_ERR = None
try:
    from datadog import api, initialize

    HAS_DATADOG = True
except Exception:
    DATADOG_IMP_ERR = traceback.format_exc()
    HAS_DATADOG = False

from ansible.module_utils.basic import AnsibleModule, missing_required_lib


def main():
    module = AnsibleModule(
        argument_spec=dict(
            api_key=dict(required=True, no_log=True),
            api_host=dict(),
            app_key=dict(required=True, no_log=True),
            state=dict(required=True, choices=["present", "absent", "mute", "unmute"]),
            type=dict(
                choices=[
                    "metric alert",
                    "service check",
                    "event alert",
                    "event-v2 alert",
                    "process alert",
                    "log alert",
                    "query alert",
                    "trace-analytics alert",
                    "rum alert",
                    "composite",
                ]
            ),
            name=dict(required=True),
            query=dict(),
            notification_message=dict(no_log=True),
            silenced=dict(type="dict"),
            notify_no_data=dict(default=False, type="bool"),
            no_data_timeframe=dict(),
            timeout_h=dict(),
            renotify_interval=dict(),
            escalation_message=dict(),
            notify_audit=dict(default=False, type="bool"),
            thresholds=dict(type="dict"),
            tags=dict(type="list", elements="str"),
            locked=dict(default=False, type="bool"),
            require_full_window=dict(type="bool"),
            new_host_delay=dict(),
            evaluation_delay=dict(),
            id=dict(),
            include_tags=dict(default=True, type="bool"),
            priority=dict(type="int"),
            notification_preset_name=dict(choices=["show_all", "hide_query", "hide_handles", "hide_all"]),
            renotify_occurrences=dict(type="int"),
            renotify_statuses=dict(type="list", elements="str", choices=["alert", "warn", "no data"]),
        )
    )

    # Prepare Datadog
    if not HAS_DATADOG:
        module.fail_json(msg=missing_required_lib("datadogpy"), exception=DATADOG_IMP_ERR)

    options = {
        "api_key": module.params["api_key"],
        "api_host": module.params["api_host"],
        "app_key": module.params["app_key"],
    }

    initialize(**options)

    # Check if api_key and app_key is correct or not
    # if not, then fail here.
    response = api.Monitor.get_all()
    if isinstance(response, dict):
        msg = response.get("errors", None)
        if msg:
            module.fail_json(msg=f"Failed to connect Datadog server using given app_key and api_key : {msg[0]}")

    if module.params["state"] == "present":
        install_monitor(module)
    elif module.params["state"] == "absent":
        delete_monitor(module)
    elif module.params["state"] == "mute":
        module.deprecate(
            "The 'mute' state is deprecated and will be removed in community.general 15.0.0. "
            "Use the community.general.datadog_downtime module to manage monitor downtimes instead.",
            version="15.0.0",
            collection_name="community.general",
        )
        mute_monitor(module)
    elif module.params["state"] == "unmute":
        module.deprecate(
            "The 'unmute' state is deprecated and will be removed in community.general 15.0.0. "
            "Use the community.general.datadog_downtime module to manage monitor downtimes instead.",
            version="15.0.0",
            collection_name="community.general",
        )
        unmute_monitor(module)


def _fix_template_vars(message):
    if message:
        return message.replace("[[", "{{").replace("]]", "}}")
    return message


def _get_monitor(module):
    if module.params["id"] is not None:
        monitor = api.Monitor.get(module.params["id"])
        if "errors" in monitor:
            module.fail_json(
                msg=f"Failed to retrieve monitor with id {module.params['id']}, errors are {monitor['errors']}"
            )
        return monitor
    else:
        monitors = api.Monitor.get_all()
        for monitor in monitors:
            if monitor["name"] == _fix_template_vars(module.params["name"]):
                return monitor
    return {}


def _post_monitor(module, options):
    try:
        kwargs = dict(
            type=module.params["type"],
            query=module.params["query"],
            name=_fix_template_vars(module.params["name"]),
            message=_fix_template_vars(module.params["notification_message"]),
            escalation_message=_fix_template_vars(module.params["escalation_message"]),
            priority=module.params["priority"],
            options=options,
        )
        if module.params["tags"] is not None:
            kwargs["tags"] = module.params["tags"]
        msg = api.Monitor.create(**kwargs)
        if "errors" in msg:
            module.fail_json(msg=str(msg["errors"]))
        else:
            module.exit_json(changed=True, msg=msg)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


def _equal_dicts(a, b, ignore_keys):
    ka = set(a).difference(ignore_keys)
    kb = set(b).difference(ignore_keys)
    return ka == kb and all(a[k] == b[k] for k in ka)


def _update_monitor(module, monitor, options):
    try:
        kwargs = dict(
            id=monitor["id"],
            query=module.params["query"],
            name=_fix_template_vars(module.params["name"]),
            message=_fix_template_vars(module.params["notification_message"]),
            escalation_message=_fix_template_vars(module.params["escalation_message"]),
            priority=module.params["priority"],
            options=options,
        )
        if module.params["tags"] is not None:
            kwargs["tags"] = module.params["tags"]
        msg = api.Monitor.update(**kwargs)

        if "errors" in msg:
            module.fail_json(msg=str(msg["errors"]))
        elif _equal_dicts(
            msg, monitor, ["creator", "overall_state", "modified", "matching_downtimes", "overall_state_modified"]
        ):
            module.exit_json(changed=False, msg=msg)
        else:
            module.exit_json(changed=True, msg=msg)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


def install_monitor(module):
    options = {
        "silenced": module.params["silenced"],
        "notify_no_data": module.boolean(module.params["notify_no_data"]),
        "no_data_timeframe": module.params["no_data_timeframe"],
        "timeout_h": module.params["timeout_h"],
        "renotify_interval": module.params["renotify_interval"],
        "escalation_message": module.params["escalation_message"],
        "notify_audit": module.boolean(module.params["notify_audit"]),
        "locked": module.boolean(module.params["locked"]),
        "require_full_window": module.params["require_full_window"],
        "new_host_delay": module.params["new_host_delay"],
        "evaluation_delay": module.params["evaluation_delay"],
        "include_tags": module.params["include_tags"],
        "notification_preset_name": module.params["notification_preset_name"],
        "renotify_occurrences": module.params["renotify_occurrences"],
        "renotify_statuses": module.params["renotify_statuses"],
    }

    if module.params["type"] == "service check":
        options["thresholds"] = module.params["thresholds"] or {"ok": 1, "critical": 1, "warning": 1}
    if (
        module.params["type"] in ["metric alert", "log alert", "query alert", "trace-analytics alert", "rum alert"]
        and module.params["thresholds"] is not None
    ):
        options["thresholds"] = module.params["thresholds"]

    monitor = _get_monitor(module)
    if not monitor:
        _post_monitor(module, options)
    else:
        _update_monitor(module, monitor, options)


def delete_monitor(module):
    monitor = _get_monitor(module)
    if not monitor:
        module.exit_json(changed=False)
    try:
        msg = api.Monitor.delete(monitor["id"])
        module.exit_json(changed=True, msg=msg)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


def mute_monitor(module):
    monitor = _get_monitor(module)
    if not monitor:
        module.fail_json(msg=f"Monitor {module.params['name']} not found!")
    elif monitor["options"]["silenced"]:
        module.fail_json(
            msg="Monitor is already muted. Datadog does not allow to modify muted alerts, consider unmuting it first."
        )
    elif (
        module.params["silenced"] is not None
        and len(set(monitor["options"]["silenced"]) ^ set(module.params["silenced"])) == 0
    ):
        module.exit_json(changed=False)
    try:
        if module.params["silenced"] is None or module.params["silenced"] == "":
            msg = api.Monitor.mute(id=monitor["id"])
        else:
            msg = api.Monitor.mute(id=monitor["id"], silenced=module.params["silenced"])
        module.exit_json(changed=True, msg=msg)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


def unmute_monitor(module):
    monitor = _get_monitor(module)
    if not monitor:
        module.fail_json(msg=f"Monitor {module.params['name']} not found!")
    elif not monitor["options"]["silenced"]:
        module.exit_json(changed=False)
    try:
        msg = api.Monitor.unmute(monitor["id"])
        module.exit_json(changed=True, msg=msg)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2017-present Alibaba Group Holding Limited. He Guimin <heguimin36@163.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
#
#  This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible. If not, see http://www.gnu.org/licenses/.

from __future__ import annotations

DOCUMENTATION = r"""
module: ali_instance
short_description: Create, Start, Stop, Restart or Terminate an Instance in ECS; Add or Remove Instance to/from a Security
  Group
description:
  - Create, start, stop, restart, modify or terminate ECS instances.
  - Add or remove ecs instances to/from security group.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  state:
    description:
      - The state of the instance after operating.
    default: 'present'
    choices: ['present', 'running', 'stopped', 'restarted', 'absent']
    type: str
  availability_zone:
    description:
      - Aliyun availability zone ID in which to launch the instance. If it is not specified, it is allocated by system automatically.
    aliases: ['alicloud_zone', 'zone_id']
    type: str
  image_id:
    description:
      - Image ID used to launch instances. Required when O(state=present) and creating new ECS instances.
    aliases: ['image']
    type: str
  instance_type:
    description:
      - Instance type used to launch instances. Required when O(state=present) and creating new ECS instances.
    aliases: ['type']
    type: str
  security_groups:
    description:
      - A list of security group IDs.
    aliases: ['group_ids']
    type: list
    elements: str
  vswitch_id:
    description:
      - The subnet ID in which to launch the instances (VPC).
    aliases: ['subnet_id']
    type: str
  instance_name:
    description:
      - The name of ECS instance, which is a string of 2 to 128 Chinese or English characters. It must begin with an uppercase/lowercase
        letter or a Chinese character and can contain numerals, V(.), V(_) or V(-). It cannot begin with V(http://) or V(https://).
    aliases: ['name']
    type: str
  description:
    description:
      - The description of ECS instance, which is a string of 2 to 256 characters. It cannot begin with V(http://) or V(https://).
    type: str
  internet_charge_type:
    description:
      - Internet charge type of ECS instance.
    default: 'PayByBandwidth'
    choices: ['PayByBandwidth', 'PayByTraffic']
    type: str
  max_bandwidth_in:
    description:
      - Maximum incoming bandwidth from the public network, measured in Mbps (Megabits per second).
    default: 200
    type: int
  max_bandwidth_out:
    description:
      - Maximum outgoing bandwidth to the public network, measured in Mbps (Megabits per second). Required when O(allocate_public_ip=true).
        Ignored when O(allocate_public_ip=false).
    default: 0
    type: int
  host_name:
    description:
      - Instance host name. Ordered hostname is not supported.
    type: str
  unique_suffix:
    description:
      - Specifies whether to add sequential suffixes to the host_name. The sequential suffix ranges from 001 to 999.
    default: false
    type: bool
    version_added: '0.2.0'
  password:
    description:
      - The password to login instance. After rebooting instances, modified password is effective.
    type: str
  system_disk_category:
    description:
      - Category of the system disk.
    default: 'cloud_efficiency'
    choices: ['cloud_efficiency', 'cloud_ssd']
    type: str
  system_disk_size:
    description:
      - Size of the system disk, in GB. The valid values are V(40)~V(500).
    default: 40
    type: int
  system_disk_name:
    description:
      - Name of the system disk.
    type: str
  system_disk_description:
    description:
      - Description of the system disk.
    type: str
  count:
    description:
      - The number of the new instance. An integer value which indicates how many instances that match O(count_tag) should
        be running. Instances are either created or terminated based on this value.
    default: 1
    type: int
  count_tag:
    description:
      - O(count) determines how many instances based on a specific tag criteria should be present. This can be expressed in
        multiple ways and is shown in the EXAMPLES section. The specified count_tag must already exist or be passed in as
        the O(tags) option. If it is not specified, it is replaced by O(instance_name).
    type: str
  allocate_public_ip:
    description:
      - Whether allocate a public IP for the new instance.
    default: false
    aliases: ['assign_public_ip']
    type: bool
  instance_charge_type:
    description:
      - The charge type of the instance.
    choices: ['PrePaid', 'PostPaid']
    default: 'PostPaid'
    type: str
  period:
    description:
      - The charge duration of the instance, in months. Required when O(instance_charge_type=PrePaid).
      - The valid value are [V(1-9), V(12), V(24), V(36)].
    default: 1
    type: int
  auto_renew:
    description:
      - Whether automate renew the charge of the instance.
    type: bool
    default: false
  auto_renew_period:
    description:
      - The duration of the automatic renew the charge of the instance. Required when O(auto_renew=true).
    choices: [1, 2, 3, 6, 12]
    type: int
  instance_ids:
    description:
      - A list of instance IDs. It is required when need to operate existing instances. If it is specified, O(count) is ignored.
    type: list
    elements: str
  force:
    description:
      - Whether the current operation needs to be execute forcibly.
    default: false
    type: bool
  tags:
    description:
      - A hash/dictionaries of instance tags, to add to the new instance or for starting/stopping instance by tag. V({"key":"value"}).
    aliases: ["instance_tags"]
    type: dict
    version_added: '0.2.0'
  purge_tags:
    description:
      - Delete any tags not specified in the task that are on the instance. If V(true), it means you have to specify all the
        desired tags on each task affecting an instance.
    default: false
    type: bool
    version_added: '0.2.0'
  key_name:
    description:
      - The name of key pair which is used to access ECS instance in SSH.
    type: str
    aliases: ['keypair']
  user_data:
    description:
      - User-defined data to customize the startup behaviors of an ECS instance and to pass data into an ECS instance. It
        only takes effect when launching the new ECS instances.
    type: str
  ram_role_name:
    description:
      - The name of the instance RAM role.
    type: str
    version_added: '0.2.0'
  spot_price_limit:
    description:
      - The maximum hourly price for the preemptible instance. This parameter supports a maximum of three decimal places and
        takes effect when the SpotStrategy parameter is set to SpotWithPriceLimit.
    type: float
    version_added: '0.2.0'
  spot_strategy:
    description:
      - The bidding mode of the pay-as-you-go instance. This parameter is valid when O(instance_charge_type=PostPaid).
    choices: ['NoSpot', 'SpotWithPriceLimit', 'SpotAsPriceGo']
    default: 'NoSpot'
    type: str
    version_added: '0.2.0'
  period_unit:
    description:
      - The duration unit that you are buying the resource. It is valid when O(instance_charge_type=PrePaid).
    choices: ['Month', 'Week']
    default: 'Month'
    type: str
    version_added: '0.2.0'
  dry_run:
    description:
      - Specifies whether to send a dry-run request.
      - If O(dry_run=true), Only a dry-run request is sent and no instance is created. The system checks whether the required
        parameters are set, and validates the request format, service permissions, and available ECS instances. If the validation
        fails, the corresponding error code is returned. If the validation succeeds, the DryRunOperation error code is returned.
      - If O(dry_run=false), a request is sent. If the validation succeeds, the instance is created.
    default: false
    type: bool
    version_added: '0.2.0'
  include_data_disks:
    description:
      - Whether to change instance disks charge type when changing instance charge type.
    default: true
    type: bool
    version_added: '0.2.0'
author:
  - "He Guimin (@xiaozhu36)"
requirements:
  - "Python >= 3.6"
  - "footmark >= 1.19.0"
extends_documentation_fragment:
  - community.general._alicloud
  - community.general._attributes
"""

EXAMPLES = r"""
# basic provisioning example vpc network
- name: Basic provisioning example
  hosts: localhost
  vars:
    alicloud_access_key: <your-alicloud-access-key-id>
    alicloud_secret_key: <your-alicloud-access-secret-key>
    alicloud_region: cn-beijing
    image: ubuntu1404_64_40G_cloudinit_20160727.raw
    instance_type: ecs.n4.small
    vswitch_id: vsw-abcd1234
    assign_public_ip: true
    max_bandwidth_out: 10
    host_name: myhost
    password: mypassword
    system_disk_category: cloud_efficiency
    system_disk_size: 100
    internet_charge_type: PayByBandwidth
    security_groups: ["sg-f2rwnfh23r"]

    instance_ids: ["i-abcd12346", "i-abcd12345"]
    force: true

  tasks:
    - name: Launch ECS instance in VPC network
      community.general.ali_instance:
        alicloud_access_key: '{{ alicloud_access_key }}'
        alicloud_secret_key: '{{ alicloud_secret_key }}'
        alicloud_region: '{{ alicloud_region }}'
        image: '{{ image }}'
        system_disk_category: '{{ system_disk_category }}'
        system_disk_size: '{{ system_disk_size }}'
        instance_type: '{{ instance_type }}'
        vswitch_id: '{{ vswitch_id }}'
        assign_public_ip: '{{ assign_public_ip }}'
        internet_charge_type: '{{ internet_charge_type }}'
        max_bandwidth_out: '{{ max_bandwidth_out }}'
        tags:
          Name: created_one
        host_name: '{{ host_name }}'
        password: '{{ password }}'

    - name: With count and count_tag to create a number of instances
      community.general.ali_instance:
        alicloud_access_key: '{{ alicloud_access_key }}'
        alicloud_secret_key: '{{ alicloud_secret_key }}'
        alicloud_region: '{{ alicloud_region }}'
        image: '{{ image }}'
        system_disk_category: '{{ system_disk_category }}'
        system_disk_size: '{{ system_disk_size }}'
        instance_type: '{{ instance_type }}'
        assign_public_ip: '{{ assign_public_ip }}'
        security_groups: '{{ security_groups }}'
        internet_charge_type: '{{ internet_charge_type }}'
        max_bandwidth_out: '{{ max_bandwidth_out }}'
        tags:
          Name: created_one
          Version: 0.1
        count: 2
        count_tag:
          Name: created_one
        host_name: '{{ host_name }}'
        password: '{{ password }}'

    - name: Start instance
      community.general.ali_instance:
        alicloud_access_key: '{{ alicloud_access_key }}'
        alicloud_secret_key: '{{ alicloud_secret_key }}'
        alicloud_region: '{{ alicloud_region }}'
        instance_ids: '{{ instance_ids }}'
        state: 'running'

    - name: Reboot instance forcibly
      ecs:
        alicloud_access_key: '{{ alicloud_access_key }}'
        alicloud_secret_key: '{{ alicloud_secret_key }}'
        alicloud_region: '{{ alicloud_region }}'
        instance_ids: '{{ instance_ids }}'
        state: 'restarted'
        force: '{{ force }}'

    - name: Add instances to an security group
      ecs:
        alicloud_access_key: '{{ alicloud_access_key }}'
        alicloud_secret_key: '{{ alicloud_secret_key }}'
        alicloud_region: '{{ alicloud_region }}'
        instance_ids: '{{ instance_ids }}'
        security_groups: '{{ security_groups }}'
"""

RETURN = r"""
instances:
  description: List of ECS instances.
  returned: always
  type: list
  elements: dict
  contains:
    availability_zone:
      description: The availability zone of the instance is in.
      returned: always
      type: str
      sample: cn-beijing-a
    block_device_mappings:
      description: Any block device mapping entries for the instance.
      returned: always
      type: complex
      contains:
        device_name:
          description: The device name exposed to the instance.
          returned: always
          type: str
          sample: /dev/xvda
        attach_time:
          description: The time stamp when the attachment initiated.
          returned: always
          type: str
          sample: "2018-06-25T04:08:26Z"
        delete_on_termination:
          description: Indicates whether the volume is deleted on instance termination.
          returned: always
          type: bool
          sample: true
        status:
          description: The attachment state.
          returned: always
          type: str
          sample: in_use
        volume_id:
          description: The ID of the cloud disk.
          returned: always
          type: str
          sample: d-2zei53pjsi117y6gf9t6
    cpu:
      description: The CPU core count of the instance.
      returned: always
      type: int
      sample: 4
    creation_time:
      description: The time the instance was created.
      returned: always
      type: str
      sample: "2018-06-25T04:08Z"
    description:
      description: The instance description.
      returned: always
      type: str
      sample: "my ansible instance"
    eip:
      description: The attribution of EIP associated with the instance.
      returned: always
      type: complex
      contains:
        allocation_id:
          description: The ID of the EIP.
          returned: always
          type: str
          sample: eip-12345
        internet_charge_type:
          description: The internet charge type of the EIP.
          returned: always
          type: str
          sample: "paybybandwidth"
        ip_address:
          description: EIP address.
          returned: always
          type: str
          sample: 42.10.2.2
    expired_time:
      description: The time the instance expires.
      returned: always
      type: str
      sample: "2099-12-31T15:59Z"
    gpu:
      description: The attribution of instance GPU.
      returned: always
      type: complex
      contains:
        amount:
          description: The count of the GPU.
          returned: always
          type: int
          sample: 0
        spec:
          description: The specification of the GPU.
          returned: always
          type: str
          sample: ""
    host_name:
      description: The host name of the instance.
      returned: always
      type: str
      sample: iZ2zewaoZ
    id:
      description: Alias of instance_id.
      returned: always
      type: str
      sample: i-abc12345
    instance_id:
      description: ECS instance resource ID.
      returned: always
      type: str
      sample: i-abc12345
    image_id:
      description: The ID of the image used to launch the instance.
      returned: always
      type: str
      sample: m-0011223344
    inner_ip_address:
      description: The inner IPv4 address of the classic instance.
      returned: always
      type: str
      sample: 10.0.0.2
    instance_charge_type:
      description: The instance charge type.
      returned: always
      type: str
      sample: PostPaid
    instance_name:
      description: The name of the instance.
      returned: always
      type: str
      sample: my-ecs
    instance_type:
      description: The instance type of the running instance.
      returned: always
      type: str
      sample: ecs.sn1ne.xlarge
    instance_type_family:
      description: The instance type family of the instance belongs.
      returned: always
      type: str
      sample: ecs.sn1ne
    internet_charge_type:
      description: The billing method of the network bandwidth.
      returned: always
      type: str
      sample: PayByBandwidth
    internet_max_bandwidth_in:
      description: Maximum incoming bandwidth from the internet network.
      returned: always
      type: int
      sample: 200
    internet_max_bandwidth_out:
      description: Maximum incoming bandwidth from the internet network.
      returned: always
      type: int
      sample: 20
    io_optimized:
      description: Indicates whether the instance is optimized for EBS I/O.
      returned: always
      type: bool
      sample: false
    memory:
      description: Memory size of the instance.
      returned: always
      type: int
      sample: 8192
    network_interfaces:
      description: One or more network interfaces for the instance.
      returned: always
      type: complex
      contains:
        mac_address:
          description: The MAC address.
          returned: always
          type: str
          sample: "00:11:22:33:44:55"
        network_interface_id:
          description: The ID of the network interface.
          returned: always
          type: str
          sample: eni-01234567
        primary_ip_address:
          description: The primary IPv4 address of the network interface within the vswitch.
          returned: always
          type: str
          sample: 10.0.0.1
    osname:
      description: The operation system name of the instance owned.
      returned: always
      type: str
      sample: CentOS
    ostype:
      description: The operation system type of the instance owned.
      returned: always
      type: str
      sample: linux
    private_ip_address:
      description: The IPv4 address of the network interface within the subnet.
      returned: always
      type: str
      sample: 10.0.0.1
    public_ip_address:
      description: The public IPv4 address assigned to the instance or eip address.
      returned: always
      type: str
      sample: 43.0.0.1
    resource_group_id:
      description: The ID of the resource group to which the instance belongs.
      returned: always
      type: str
      sample: my-ecs-group
    security_groups:
      description: One or more security groups for the instance.
      returned: always
      type: list
      elements: dict
      contains:
        group_id:
          description: The ID of the security group.
          returned: always
          type: str
          sample: sg-0123456
        group_name:
          description: The name of the security group.
          returned: always
          type: str
          sample: my-security-group
    status:
      description: The current status of the instance.
      returned: always
      type: str
      sample: running
    tags:
      description: Any tags assigned to the instance.
      returned: always
      type: dict
      sample:
    user_data:
      description: User-defined data.
      returned: always
      type: dict
      sample:
    vswitch_id:
      description: The ID of the vswitch in which the instance is running.
      returned: always
      type: str
      sample: vsw-dew00abcdef
    vpc_id:
      description: The ID of the VPC the instance is in.
      returned: always
      type: str
      sample: vpc-0011223344
    spot_price_limit:
      description:
        - The maximum hourly price for the preemptible instance.
      returned: always
      type: float
      sample: 0.97
    spot_strategy:
      description:
        - The bidding mode of the pay-as-you-go instance.
      returned: always
      type: str
      sample: NoSpot
ids:
  description: List of ECS instance IDs.
  returned: always
  type: list
  sample: ["i-12345er", "i-3245fs"]
"""

import re
import time

from ansible.module_utils.basic import AnsibleModule, missing_required_lib

from ansible_collections.community.general.plugins.module_utils._alicloud_ecs import (
    FOOTMARK_IMP_ERR,
    HAS_FOOTMARK,
    ecs_argument_spec,
    ecs_connect,
)


def get_instances_info(connection, ids):
    result = []
    instances = connection.describe_instances(instance_ids=ids)
    if len(instances) > 0:
        for inst in instances:
            volumes = connection.describe_disks(instance_id=inst.id)
            inst.block_device_mappings = volumes
            inst.user_data = inst.describe_user_data()
            result.append(inst.read())
    return result


def run_instance(module, ecs, exact_count):
    if exact_count <= 0:
        return None
    zone_id = module.params["availability_zone"]
    image_id = module.params["image_id"]
    instance_type = module.params["instance_type"]
    security_groups = module.params["security_groups"]
    vswitch_id = module.params["vswitch_id"]
    instance_name = module.params["instance_name"]
    description = module.params["description"]
    internet_charge_type = module.params["internet_charge_type"]
    max_bandwidth_out = module.params["max_bandwidth_out"]
    max_bandwidth_in = module.params["max_bandwidth_in"]
    host_name = module.params["host_name"]
    password = module.params["password"]
    system_disk_category = module.params["system_disk_category"]
    system_disk_size = module.params["system_disk_size"]
    system_disk_name = module.params["system_disk_name"]
    system_disk_description = module.params["system_disk_description"]
    # allocate_public_ip = module.params["allocate_public_ip"]  TODO - this is unused!
    period = module.params["period"]
    auto_renew = module.params["auto_renew"]
    instance_charge_type = module.params["instance_charge_type"]
    auto_renew_period = module.params["auto_renew_period"]
    user_data = module.params["user_data"]
    key_name = module.params["key_name"]
    ram_role_name = module.params["ram_role_name"]
    spot_price_limit = module.params["spot_price_limit"]
    spot_strategy = module.params["spot_strategy"]
    unique_suffix = module.params["unique_suffix"]
    # check whether the required parameter passed or not
    if not image_id:
        module.fail_json(msg="image_id is required for new instance")
    if not instance_type:
        module.fail_json(msg="instance_type is required for new instance")
    if not isinstance(security_groups, list):
        module.fail_json(msg="The parameter security_groups should be a list, aborting")
    if len(security_groups) <= 0:
        module.fail_json(
            msg="Expected the parameter security_groups is non-empty when create new ECS instances, aborting"
        )

    client_token = f"Ansible-Alicloud-{hash(str(module.params))}-{time.time()}"

    try:
        # call to create_instance method from footmark
        instances = ecs.run_instances(
            image_id=image_id,
            instance_type=instance_type,
            security_group_id=security_groups[0],
            zone_id=zone_id,
            instance_name=instance_name,
            description=description,
            internet_charge_type=internet_charge_type,
            internet_max_bandwidth_out=max_bandwidth_out,
            internet_max_bandwidth_in=max_bandwidth_in,
            host_name=host_name,
            password=password,
            io_optimized="optimized",
            system_disk_category=system_disk_category,
            system_disk_size=system_disk_size,
            system_disk_disk_name=system_disk_name,
            system_disk_description=system_disk_description,
            vswitch_id=vswitch_id,
            amount=exact_count,
            instance_charge_type=instance_charge_type,
            period=period,
            period_unit="Month",
            auto_renew=auto_renew,
            auto_renew_period=auto_renew_period,
            key_pair_name=key_name,
            user_data=user_data,
            client_token=client_token,
            ram_role_name=ram_role_name,
            spot_price_limit=spot_price_limit,
            spot_strategy=spot_strategy,
            unique_suffix=unique_suffix,
        )

    except Exception as e:
        module.fail_json(msg=f"Unable to create instance, error: {e}")

    return instances


def modify_instance(module, instance):
    # According to state to modify instance's some special attribute
    state = module.params["state"]
    name = module.params["instance_name"]
    unique_suffix = module.params["unique_suffix"]
    if not name:
        name = instance.name

    description = module.params["description"]
    if not description:
        description = instance.description

    host_name = module.params["host_name"]
    if unique_suffix and host_name:
        suffix = instance.host_name[-3:]
        host_name = host_name + suffix

    if not host_name:
        host_name = instance.host_name

    # password can be modified only when restart instance
    password = ""
    if state == "restarted":
        password = module.params["password"]

    # userdata can be modified only when instance is stopped
    instance.user_data = instance.describe_user_data()
    user_data = instance.user_data
    if state == "stopped":
        user_data = module.params["user_data"].encode()

    try:
        return instance.modify(
            name=name, description=description, host_name=host_name, password=password, user_data=user_data
        )
    except Exception as e:
        module.fail_json(msg=f"Modify instance {instance.id} attribute got an error: {e}")


def wait_for_instance_modify_charge(ecs, instance_ids, charge_type, delay=10, timeout=300):
    """
    To verify instance charge type has become expected after modify instance charge type
    """
    try:
        while True:
            instances = ecs.describe_instances(instance_ids=instance_ids)
            flag = True
            for inst in instances:
                if inst and inst.instance_charge_type != charge_type:
                    flag = False
            if flag:
                return
            timeout -= delay
            time.sleep(delay)
            if timeout <= 0:
                raise Exception(f"Timeout Error: Waiting for instance to {charge_type}. ")
    except Exception as e:
        raise e


def main():
    argument_spec = ecs_argument_spec()
    argument_spec.update(
        dict(
            security_groups=dict(type="list", elements="str", aliases=["group_ids"]),
            availability_zone=dict(type="str", aliases=["alicloud_zone", "zone_id"]),
            instance_type=dict(type="str", aliases=["type"]),
            image_id=dict(type="str", aliases=["image"]),
            count=dict(type="int", default=1),
            count_tag=dict(type="str"),
            vswitch_id=dict(type="str", aliases=["subnet_id"]),
            instance_name=dict(type="str", aliases=["name"]),
            host_name=dict(type="str"),
            password=dict(type="str", no_log=True),
            internet_charge_type=dict(type="str", default="PayByBandwidth", choices=["PayByBandwidth", "PayByTraffic"]),
            max_bandwidth_in=dict(type="int", default=200),
            max_bandwidth_out=dict(type="int", default=0),
            system_disk_category=dict(
                type="str", default="cloud_efficiency", choices=["cloud_efficiency", "cloud_ssd"]
            ),
            system_disk_size=dict(type="int", default=40),
            system_disk_name=dict(type="str"),
            system_disk_description=dict(type="str"),
            force=dict(type="bool", default=False),
            tags=dict(type="dict", aliases=["instance_tags"]),
            purge_tags=dict(type="bool", default=False),
            state=dict(default="present", choices=["present", "running", "stopped", "restarted", "absent"]),
            description=dict(type="str"),
            allocate_public_ip=dict(type="bool", aliases=["assign_public_ip"], default=False),
            instance_charge_type=dict(type="str", default="PostPaid", choices=["PrePaid", "PostPaid"]),
            period=dict(type="int", default=1),
            auto_renew=dict(type="bool", default=False),
            instance_ids=dict(type="list", elements="str"),
            auto_renew_period=dict(type="int", choices=[1, 2, 3, 6, 12]),
            key_name=dict(type="str", aliases=["keypair"]),
            user_data=dict(type="str"),
            ram_role_name=dict(type="str"),
            spot_price_limit=dict(type="float"),
            spot_strategy=dict(type="str", default="NoSpot", choices=["NoSpot", "SpotWithPriceLimit", "SpotAsPriceGo"]),
            unique_suffix=dict(type="bool", default=False),
            period_unit=dict(type="str", default="Month", choices=["Month", "Week"]),
            dry_run=dict(type="bool", default=False),
            include_data_disks=dict(type="bool", default=True),
        )
    )
    module = AnsibleModule(argument_spec=argument_spec)

    if HAS_FOOTMARK is False:
        module.fail_json(msg=missing_required_lib("footmark"), exception=FOOTMARK_IMP_ERR)

    ecs = ecs_connect(module)
    host_name = module.params["host_name"]
    state = module.params["state"]
    instance_ids = module.params["instance_ids"]
    count_tag = module.params["count_tag"]
    count = module.params["count"]
    instance_name = module.params["instance_name"]
    force = module.params["force"]
    zone_id = module.params["availability_zone"]
    key_name = module.params["key_name"]
    tags = module.params["tags"]
    max_bandwidth_out = module.params["max_bandwidth_out"]
    instance_charge_type = module.params["instance_charge_type"]
    if instance_charge_type == "PrePaid":
        module.params["spot_strategy"] = ""
    changed = False

    instances = []
    if instance_ids:
        if not isinstance(instance_ids, list):
            module.fail_json(msg="The parameter instance_ids should be a list, aborting")
        instances = ecs.describe_instances(zone_id=zone_id, instance_ids=instance_ids)
        if not instances:
            module.fail_json(
                msg=f"There are no instances in our record based on instance_ids {instance_ids}. Please check it and try again."
            )
    elif count_tag:
        instances = ecs.describe_instances(zone_id=zone_id, tags=eval(count_tag))
    elif instance_name:
        instances = ecs.describe_instances(zone_id=zone_id, instance_name=instance_name)

    ids = []
    if state == "absent":
        if len(instances) < 1:
            module.fail_json(
                msg="Please specify ECS instances that you want to operate by using "
                "parameters instance_ids, tags or instance_name, aborting"
            )
        try:
            targets = []
            for inst in instances:
                if inst.status != "stopped" and not force:
                    module.fail_json(msg="Instance is running, and please stop it or set 'force' as True.")
                targets.append(inst.id)
            if ecs.delete_instances(instance_ids=targets, force=force):
                changed = True
                ids.extend(targets)

            module.exit_json(changed=changed, ids=ids, instances=[])
        except Exception as e:
            module.fail_json(msg=f"Delete instance got an error: {e}")

    if module.params["allocate_public_ip"] and max_bandwidth_out < 0:
        module.fail_json(msg="'max_bandwidth_out' should be greater than 0 when 'allocate_public_ip' is True.")
    if not module.params["allocate_public_ip"]:
        module.params["max_bandwidth_out"] = 0

    if state == "present":
        if not instance_ids:
            if len(instances) > count:
                for _i in range(0, len(instances) - count):
                    inst = instances[len(instances) - 1]
                    if inst.status != "stopped" and not force:
                        module.fail_json(
                            msg=f"That to delete instance {inst.id} is failed results from it is running, "
                            "and please stop it or set 'force' as True."
                        )
                    try:
                        if inst.terminate(force=force):
                            changed = True
                    except Exception as e:
                        module.fail_json(msg=f"Delete instance {inst.id} got an error: {e}")
                    instances.pop(len(instances) - 1)
            else:
                try:
                    if re.search(r"-\[\d+,\d+\]-", host_name):
                        module.fail_json(
                            msg="Ordered hostname is not supported, If you want to add an ordered "
                            "suffix to the hostname, you can set unique_suffix to True"
                        )
                    new_instances = run_instance(module, ecs, count - len(instances))
                    if new_instances:
                        changed = True
                        instances.extend(new_instances)
                except Exception as e:
                    module.fail_json(msg=f"Create new instances got an error: {e}")

        # Security Group join/leave begin
        security_groups = module.params["security_groups"]
        if security_groups:
            if not isinstance(security_groups, list):
                module.fail_json(msg="The parameter security_groups should be a list, aborting")
            for inst in instances:
                existing = inst.security_group_ids["security_group_id"]
                remove = list(set(existing).difference(set(security_groups)))
                add = list(set(security_groups).difference(set(existing)))
                for sg in remove:
                    if inst.leave_security_group(sg):
                        changed = True
                for sg in add:
                    if inst.join_security_group(sg):
                        changed = True
        # Security Group join/leave ends here

        # Attach/Detach key pair
        inst_ids = []
        for inst in instances:
            if key_name is not None and key_name != inst.key_name:
                if key_name == "":
                    if inst.detach_key_pair():
                        changed = True
                else:
                    inst_ids.append(inst.id)
        if inst_ids:
            changed = ecs.attach_key_pair(instance_ids=inst_ids, key_pair_name=key_name)

        # Modify instance attribute
        for inst in instances:
            if modify_instance(module, inst):
                changed = True
            if inst.id not in ids:
                ids.append(inst.id)

        # Modify instance charge type
        ids = []
        for inst in instances:
            if inst.instance_charge_type != instance_charge_type:
                ids.append(inst.id)
        if ids:
            params = {
                "instance_ids": ids,
                "instance_charge_type": instance_charge_type,
                "include_data_disks": module.params["include_data_disks"],
                "dry_run": module.params["dry_run"],
                "auto_pay": True,
            }
            if instance_charge_type == "PrePaid":
                params["period"] = module.params["period"]
                params["period_unit"] = module.params["period_unit"]

            if ecs.modify_instance_charge_type(**params):
                changed = True
                wait_for_instance_modify_charge(ecs, ids, instance_charge_type)

    else:
        if len(instances) < 1:
            module.fail_json(
                msg="Please specify ECS instances that you want to operate by using "
                "parameters instance_ids, tags or instance_name, aborting"
            )
        if state == "running":
            try:
                targets = []
                for inst in instances:
                    if modify_instance(module, inst):
                        changed = True
                    if inst.status != "running":
                        targets.append(inst.id)
                    ids.append(inst.id)
                if targets and ecs.start_instances(instance_ids=targets):
                    changed = True
                    ids.extend(targets)
            except Exception as e:
                module.fail_json(msg=f"Start instances got an error: {e}")
        elif state == "stopped":
            try:
                targets = []
                for inst in instances:
                    if inst.status != "stopped":
                        targets.append(inst.id)
                if targets and ecs.stop_instances(instance_ids=targets, force_stop=force):
                    changed = True
                    ids.extend(targets)
                for inst in instances:
                    if modify_instance(module, inst):
                        changed = True
            except Exception as e:
                module.fail_json(msg=f"Stop instances got an error: {e}")
        elif state == "restarted":
            try:
                targets = []
                for inst in instances:
                    if modify_instance(module, inst):
                        changed = True
                        targets.append(inst.id)
                if ecs.reboot_instances(instance_ids=targets, force_stop=module.params["force"]):
                    changed = True
                    ids.extend(targets)
            except Exception as e:
                module.fail_json(msg=f"Reboot instances got an error: {e}")

    tags = module.params["tags"]
    if module.params["purge_tags"]:
        for inst in instances:
            if not tags:
                tags = inst.tags
            try:
                if inst.remove_tags(tags):
                    changed = True
            except Exception as e:
                module.fail_json(msg=f"{e}")
        module.exit_json(changed=changed, instances=get_instances_info(ecs, ids))

    if tags:
        for inst in instances:
            try:
                if inst.add_tags(tags):
                    changed = True
            except Exception as e:
                module.fail_json(msg=f"{e}")
    module.exit_json(changed=changed, instances=get_instances_info(ecs, ids))


if __name__ == "__main__":
    main()



# This code is part of the Ansible collection community.docker, but is an independent component.
# This particular file, and this file only, is based on the Docker SDK for Python (https://github.com/docker/docker-py/)
#
# Copyright (c) 2016-2022 Docker, Inc.
#
# It is licensed under the Apache 2.0 license (see LICENSES/Apache-2.0.txt in this collection)
# SPDX-License-Identifier: Apache-2.0

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t
from queue import Empty

from .. import constants
from .._import_helper import HTTPAdapter, urllib3, urllib3_connection
from .basehttpadapter import BaseHTTPAdapter
from .npipesocket import NpipeSocket

if t.TYPE_CHECKING:
    from collections.abc import Mapping

    from requests import PreparedRequest


RecentlyUsedContainer = urllib3._collections.RecentlyUsedContainer


class NpipeHTTPConnection(urllib3_connection.HTTPConnection):
    def __init__(self, npipe_path: str, timeout: int | float = 60) -> None:
        super().__init__("localhost", timeout=timeout)
        self.npipe_path = npipe_path
        self.timeout = timeout

    def connect(self) -> None:
        sock = NpipeSocket()
        sock.settimeout(self.timeout)
        sock.connect(self.npipe_path)
        self.sock = sock


class NpipeHTTPConnectionPool(urllib3.connectionpool.HTTPConnectionPool):
    def __init__(
        self, npipe_path: str, timeout: int | float = 60, maxsize: int = 10
    ) -> None:
        super().__init__("localhost", timeout=timeout, maxsize=maxsize)
        self.npipe_path = npipe_path
        self.timeout = timeout

    def _new_conn(self) -> NpipeHTTPConnection:
        return NpipeHTTPConnection(self.npipe_path, self.timeout)

    # When re-using connections, urllib3 tries to call select() on our
    # NpipeSocket instance, causing a crash. To circumvent this, we override
    # _get_conn, where that check happens.
    def _get_conn(self, timeout: int | float) -> NpipeHTTPConnection:
        conn = None
        try:
            conn = self.pool.get(block=self.block, timeout=timeout)

        except AttributeError as exc:  # self.pool is None
            raise urllib3.exceptions.ClosedPoolError(self, "Pool is closed.") from exc

        except Empty as exc:
            if self.block:
                raise urllib3.exceptions.EmptyPoolError(
                    self,
                    "Pool reached maximum size and no more connections are allowed.",
                ) from exc
            # Oh well, we'll create a new connection then

        return conn or self._new_conn()


class NpipeHTTPAdapter(BaseHTTPAdapter):
    __attrs__ = HTTPAdapter.__attrs__ + [
        "npipe_path",
        "pools",
        "timeout",
        "max_pool_size",
    ]

    def __init__(
        self,
        base_url: str,
        timeout: int | float = 60,
        pool_connections: int = constants.DEFAULT_NUM_POOLS,
        max_pool_size: int = constants.DEFAULT_MAX_POOL_SIZE,
    ) -> None:
        self.npipe_path = base_url.replace("npipe://", "")
        self.timeout = timeout
        self.max_pool_size = max_pool_size
        self.pools = RecentlyUsedContainer(
            pool_connections, dispose_func=lambda p: p.close()
        )
        super().__init__()

    def get_connection(
        self, url: str | bytes, proxies: Mapping[str, str] | None = None
    ) -> NpipeHTTPConnectionPool:
        with self.pools.lock:
            pool = self.pools.get(url)
            if pool:
                return pool

            pool = NpipeHTTPConnectionPool(
                self.npipe_path, self.timeout, maxsize=self.max_pool_size
            )
            self.pools[url] = pool

        return pool

    def request_url(
        self, request: PreparedRequest, proxies: Mapping[str, str] | None
    ) -> str:
        # The select_proxy utility in requests errors out when the provided URL
        # does not have a hostname, like is the case when using a UNIX socket.
        # Since proxies are an irrelevant notion in the case of UNIX sockets
        # anyway, we simply return the path URL directly.
        # See also: https://github.com/docker/docker-sdk-python/issues/811
        return request.path_url



#!/usr/bin/python

# Copyright 2012 Dag Wieers <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: hpilo_info
author: Dag Wieers (@dagwieers)
short_description: Gather information through an HP iLO interface
description:
  - This module gathers information on a specific system using its HP iLO interface. These information includes hardware and
    network related information useful for provisioning (for example macaddress, uuid).
  - This module requires the C(hpilo) python module.
extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.info_module
options:
  host:
    description:
      - The HP iLO hostname/address that is linked to the physical system.
    type: str
    required: true
  login:
    description:
      - The login name to authenticate to the HP iLO interface.
    type: str
    default: Administrator
  password:
    description:
      - The password to authenticate to the HP iLO interface.
    type: str
    default: admin
  ssl_version:
    description:
      - Change the ssl_version used.
    default: TLSv1
    type: str
    choices: ["SSLv3", "SSLv23", "TLSv1", "TLSv1_1", "TLSv1_2"]
requirements:
  - hpilo
notes:
  - This module ought to be run from a system that can access the HP iLO interface directly, either by using C(local_action)
    or using C(delegate_to).
"""

EXAMPLES = r"""
- name: Gather facts from a HP iLO interface only if the system is an HP server
  community.general.hpilo_info:
    host: YOUR_ILO_ADDRESS
    login: YOUR_ILO_LOGIN
    password: YOUR_ILO_PASSWORD
  when: cmdb_hwmodel.startswith('HP ')
  delegate_to: localhost
  register: results

- ansible.builtin.fail:
    msg: 'CMDB serial ({{ cmdb_serialno }}) does not match hardware serial ({{ results.hw_system_serial }}) !'
  when: cmdb_serialno != results.hw_system_serial
"""

RETURN = r"""
# Typical output of HP iLO_info for a physical system
hw_bios_date:
  description: BIOS date.
  returned: always
  type: str
  sample: 05/05/2011

hw_bios_version:
  description: BIOS version.
  returned: always
  type: str
  sample: P68

hw_ethX:
  description: Interface information (for each interface).
  returned: always
  type: dict
  sample:
    - macaddress: 00:11:22:33:44:55
      macaddress_dash: 00-11-22-33-44-55

hw_eth_ilo:
  description: Interface information (for the iLO network interface).
  returned: always
  type: dict
  sample:
    - macaddress: 00:11:22:33:44:BA
    - macaddress_dash: 00-11-22-33-44-BA

hw_product_name:
  description: Product name.
  returned: always
  type: str
  sample: ProLiant DL360 G7

hw_product_uuid:
  description: Product UUID.
  returned: always
  type: str
  sample: ef50bac8-2845-40ff-81d9-675315501dac

hw_system_serial:
  description: System serial number.
  returned: always
  type: str
  sample: ABC12345D6

hw_uuid:
  description: Hardware UUID.
  returned: always
  type: str
  sample: 123456ABC78901D2

host_power_status:
  description:
    - Power status of host.
    - It is one of V(ON), V(OFF) and V(UNKNOWN).
  returned: always
  type: str
  sample: "ON"
  version_added: 3.5.0
"""

import re
import traceback
import warnings

HPILO_IMP_ERR = None
try:
    import hpilo

    HAS_HPILO = True
except ImportError:
    HPILO_IMP_ERR = traceback.format_exc()
    HAS_HPILO = False

from ansible.module_utils.basic import AnsibleModule, missing_required_lib

# Suppress warnings from hpilo
warnings.simplefilter("ignore")


def parse_flat_interface(entry, non_numeric="hw_eth_ilo"):
    try:
        infoname = f"hw_eth{int(entry['Port']) - 1}"
    except Exception:
        infoname = non_numeric

    info = {"macaddress": entry["MAC"].replace("-", ":"), "macaddress_dash": entry["MAC"]}
    return (infoname, info)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            host=dict(type="str", required=True),
            login=dict(type="str", default="Administrator"),
            password=dict(type="str", default="admin", no_log=True),
            ssl_version=dict(type="str", default="TLSv1", choices=["SSLv3", "SSLv23", "TLSv1", "TLSv1_1", "TLSv1_2"]),
        ),
        supports_check_mode=True,
    )

    if not HAS_HPILO:
        module.fail_json(msg=missing_required_lib("python-hpilo"), exception=HPILO_IMP_ERR)

    host = module.params["host"]
    login = module.params["login"]
    password = module.params["password"]
    ssl_version = getattr(hpilo.ssl, f"PROTOCOL_{module.params.get('ssl_version').upper().replace('V', 'v')}")

    ilo = hpilo.Ilo(host, login=login, password=password, ssl_version=ssl_version)

    info = {
        "module_hw": True,
    }

    # TODO: Count number of CPUs, DIMMs and total memory
    try:
        data = ilo.get_host_data()
        power_state = ilo.get_host_power_status()
    except hpilo.IloCommunicationError as e:
        module.fail_json(msg=f"{e}")

    for entry in data:
        if "type" not in entry:
            continue
        elif entry["type"] == 0:  # BIOS Information
            info["hw_bios_version"] = entry["Family"]
            info["hw_bios_date"] = entry["Date"]
        elif entry["type"] == 1:  # System Information
            info["hw_uuid"] = entry["UUID"]
            info["hw_system_serial"] = entry["Serial Number"].rstrip()
            info["hw_product_name"] = entry["Product Name"]
            info["hw_product_uuid"] = entry["cUUID"]
        elif entry["type"] == 209:  # Embedded NIC MAC Assignment
            if "fields" in entry:
                for name, value in [(e["name"], e["value"]) for e in entry["fields"]]:
                    if name.startswith("Port"):
                        try:
                            infoname = f"hw_eth{int(value) - 1}"
                        except Exception:
                            infoname = "hw_eth_ilo"
                    elif name.startswith("MAC"):
                        info[infoname] = {"macaddress": value.replace("-", ":"), "macaddress_dash": value}
            else:
                (infoname, entry_info) = parse_flat_interface(entry, "hw_eth_ilo")
                info[infoname] = entry_info
        elif entry["type"] == 209:  # HPQ NIC iSCSI MAC Info
            for name, value in [(e["name"], e["value"]) for e in entry["fields"]]:
                if name.startswith("Port"):
                    try:
                        infoname = f"hw_iscsi{int(value) - 1}"
                    except Exception:
                        infoname = "hw_iscsi_ilo"
                elif name.startswith("MAC"):
                    info[infoname] = {"macaddress": value.replace("-", ":"), "macaddress_dash": value}
        elif entry["type"] == 233:  # Embedded NIC MAC Assignment (Alternate data format)
            (infoname, entry_info) = parse_flat_interface(entry, "hw_eth_ilo")
            info[infoname] = entry_info

    # Collect health (RAM/CPU data)
    health = ilo.get_embedded_health()
    info["hw_health"] = health

    memory_details_summary = health.get("memory", {}).get("memory_details_summary")
    # RAM as reported by iLO 2.10 on ProLiant BL460c Gen8
    if memory_details_summary:
        info["hw_memory_details_summary"] = memory_details_summary
        info["hw_memory_total"] = 0
        for details in memory_details_summary.values():
            cpu_total_memory_size = details.get("total_memory_size")
            if cpu_total_memory_size:
                ram = re.search(r"(\d+)\s+(\w+)", cpu_total_memory_size)
                if ram:
                    if ram.group(2) == "GB":
                        info["hw_memory_total"] = info["hw_memory_total"] + int(ram.group(1))

        # reformat into a text friendly format
        info["hw_memory_total"] = f"{info['hw_memory_total']} GB"

    # Report host state
    info["host_power_status"] = power_state or "UNKNOWN"

    module.exit_json(**info)


if __name__ == "__main__":
    main()



# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: nmap
short_description: Uses nmap to find hosts to target
description:
  - Uses a YAML configuration file with a valid YAML extension.
extends_documentation_fragment:
  - ansible.builtin.constructed
  - ansible.builtin.inventory_cache
requirements:
  - nmap CLI installed
options:
  plugin:
    description: Token that ensures this is a source file for the P(community.general.nmap#inventory) plugin.
    type: string
    required: true
    choices: ['nmap', 'community.general.nmap']
  sudo:
    description: Set to V(true) to execute a C(sudo nmap) plugin scan.
    version_added: 4.8.0
    default: false
    type: boolean
  address:
    description: Network IP or range of IPs to scan, you can use a simple range (10.2.2.15-25) or CIDR notation.
    type: string
    required: true
    env:
      - name: ANSIBLE_NMAP_ADDRESS
        version_added: 6.6.0
  exclude:
    description:
      - List of addresses to exclude.
      - For example V(10.2.2.15-25) or V(10.2.2.15,10.2.2.16).
    type: list
    elements: string
    env:
      - name: ANSIBLE_NMAP_EXCLUDE
        version_added: 6.6.0
  port:
    description:
      - Only scan specific port or port range (C(-p)).
      - For example, you could pass V(22) for a single port, V(1-65535) for a range of ports, or V(U:53,137,T:21-25,139,8080,S:9)
        to check port 53 with UDP, ports 21-25 with TCP, port 9 with SCTP, and ports 137, 139, and 8080 with all.
    type: string
    version_added: 6.5.0
  ports:
    description: Enable/disable scanning ports.
    type: boolean
    default: true
  ipv4:
    description: Use IPv4 type addresses.
    type: boolean
    default: true
  ipv6:
    description: Use IPv6 type addresses.
    type: boolean
    default: true
  udp_scan:
    description:
      - Scan using UDP.
      - Depending on your system you might need O(sudo=true) for this to work.
    type: boolean
    default: false
    version_added: 6.1.0
  icmp_timestamp:
    description:
      - Scan using ICMP Timestamp (C(-PP)).
      - Depending on your system you might need O(sudo=true) for this to work.
    type: boolean
    default: false
    version_added: 6.1.0
  open:
    description: Only scan for open (or possibly open) ports.
    type: boolean
    default: false
    version_added: 6.5.0
  dns_resolve:
    description: Whether to always (V(true)) or never (V(false)) do DNS resolution.
    type: boolean
    default: false
    version_added: 6.1.0
  dns_servers:
    description: Specify which DNS servers to use for name resolution.
    type: list
    elements: string
    version_added: 10.5.0
  use_arp_ping:
    description: Whether to always (V(true)) use the quick ARP ping or (V(false)) a slower but more reliable method.
    type: boolean
    default: true
    version_added: 7.4.0
  skip_host_discovery:
    description:
      - Skip nmap host discovery phase and treat all hosts as online (C(-Pn)).
      - Useful when scanning remote hosts over VPN or through firewalls where nmap's default discovery probes
        (TCP SYN to ports 80/443) are blocked but the target port is open.
      - When V(false) (default), nmap performs host discovery before port scanning, which may send packets
        to ports 80 and 443 regardless of the O(port) setting.
    type: boolean
    default: false
    version_added: 13.0.0
  set_name_variable:
    description:
      - Set the C(name) variable for each host.
      - When V(true), sets the C(name) variable which may trigger a warning about using a reserved name.
      - Set to V(false) to avoid the warning when C(name) is not needed as a variable.
    type: boolean
    default: true
    version_added: 13.0.0
notes:
  - At least one of O(ipv4) or O(ipv6) is required to be V(true); both can be V(true), but they cannot both be V(false).
  - 'TODO: add OS fingerprinting.'
"""
EXAMPLES = r"""
---
# inventory.config file in YAML format
plugin: community.general.nmap
strict: false
address: 192.168.0.0/24

---
# a sudo nmap scan to fully use nmap scan power.
plugin: community.general.nmap
sudo: true
strict: false
address: 192.168.0.0/24

---
# an nmap scan specifying ports and classifying results to an inventory group
plugin: community.general.nmap
address: 192.168.0.0/24
exclude: 192.168.0.1, web.example.com
port: 22, 443
groups:
  web_servers: "ports | selectattr('port', 'equalto', '443')"

---
# an nmap scan without setting the 'name' variable to avoid warnings
plugin: community.general.nmap
address: 192.168.0.0/24
set_name_variable: false
"""

import os
import re
from subprocess import PIPE, Popen

from ansible import constants as C
from ansible.errors import AnsibleParserError
from ansible.module_utils.common.process import get_bin_path
from ansible.module_utils.common.text.converters import to_native, to_text
from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable, Constructable

from ansible_collections.community.general.plugins.plugin_utils._unsafe import make_unsafe


class InventoryModule(BaseInventoryPlugin, Constructable, Cacheable):
    NAME = "community.general.nmap"
    find_host = re.compile(r"^Nmap scan report for ([\w,.,-]+)(?: \(([\w,.,:,\[,\]]+)\))?")
    find_port = re.compile(r"^(\d+)/(\w+)\s+(\w+)\s+(\w+)")

    def __init__(self):
        self._nmap = None
        super().__init__()

    def _populate(self, hosts):
        # Use constructed if applicable
        strict = self.get_option("strict")
        set_name_variable = self.get_option("set_name_variable")

        for host in hosts:
            host = make_unsafe(host)
            hostname = host["name"]
            self.inventory.add_host(hostname)
            for var, value in host.items():
                if var == "name" and not set_name_variable:
                    continue
                self.inventory.set_variable(hostname, var, value)

            # Composed variables
            self._set_composite_vars(self.get_option("compose"), host, hostname, strict=strict)

            # Complex groups based on jinja2 conditionals, hosts that meet the conditional are added to group
            self._add_host_to_composed_groups(self.get_option("groups"), host, hostname, strict=strict)

            # Create groups based on variable values and add the corresponding hosts to it
            self._add_host_to_keyed_groups(self.get_option("keyed_groups"), host, hostname, strict=strict)

    def verify_file(self, path):
        valid = False
        if super().verify_file(path):
            file_name, ext = os.path.splitext(path)

            if not ext or ext in C.YAML_FILENAME_EXTENSIONS:
                valid = True

        return valid

    def parse(self, inventory, loader, path, cache=True):
        try:
            self._nmap = get_bin_path("nmap")
        except ValueError as e:
            raise AnsibleParserError(f"nmap inventory plugin requires the nmap cli tool to work: {e}") from e

        super().parse(inventory, loader, path, cache=cache)

        self._read_config_data(path)

        cache_key = self.get_cache_key(path)

        # cache may be True or False at this point to indicate if the inventory is being refreshed
        # get the user's cache option too to see if we should save the cache if it is changing
        user_cache_setting = self.get_option("cache")

        # read if the user has caching enabled and the cache isn't being refreshed
        attempt_to_read_cache = user_cache_setting and cache
        # update if the user has caching enabled and the cache is being refreshed; update this value to True if the cache has expired below
        cache_needs_update = user_cache_setting and not cache

        if attempt_to_read_cache:
            try:
                results = self._cache[cache_key]
            except KeyError:
                # This occurs if the cache_key is not in the cache or if the cache_key expired, so the cache needs to be updated
                cache_needs_update = True

        if not user_cache_setting or cache_needs_update:
            # setup command
            cmd = [self._nmap]

            if self.get_option("sudo"):
                cmd.insert(0, "sudo")

            if self.get_option("port"):
                cmd.append("-p")
                cmd.append(self.get_option("port"))

            if not self.get_option("ports"):
                cmd.append("-sP")

            if self.get_option("ipv4") and not self.get_option("ipv6"):
                cmd.append("-4")
            elif self.get_option("ipv6") and not self.get_option("ipv4"):
                cmd.append("-6")
            elif not self.get_option("ipv6") and not self.get_option("ipv4"):
                raise AnsibleParserError("One of ipv4 or ipv6 must be enabled for this plugin")

            if self.get_option("exclude"):
                cmd.append("--exclude")
                cmd.append(",".join(self.get_option("exclude")))

            if self.get_option("dns_resolve"):
                cmd.append("-n")

            if self.get_option("dns_servers"):
                cmd.append("--dns-servers")
                cmd.append(",".join(self.get_option("dns_servers")))

            if self.get_option("udp_scan"):
                cmd.append("-sU")

            if self.get_option("icmp_timestamp"):
                cmd.append("-PP")

            if self.get_option("open"):
                cmd.append("--open")

            if not self.get_option("use_arp_ping"):
                cmd.append("--disable-arp-ping")

            if self.get_option("skip_host_discovery"):
                cmd.append("-Pn")

            cmd.append(self.get_option("address"))
            try:
                # execute
                p = Popen(cmd, stdout=PIPE, stderr=PIPE)
                stdout, stderr = p.communicate()
                if p.returncode != 0:
                    raise AnsibleParserError(f"Failed to run nmap, rc={p.returncode}: {to_native(stderr)}")

                # parse results
                host = None
                ip = None
                ports = []
                results = []

                try:
                    t_stdout = to_text(stdout, errors="surrogate_or_strict")
                except UnicodeError as e:
                    raise AnsibleParserError(f"Invalid (non unicode) input returned: {e}") from e

                for line in t_stdout.splitlines():
                    hits = self.find_host.match(line)
                    if hits:
                        if host is not None and ports:
                            results[-1]["ports"] = ports

                        # if dns only shows arpa, just use ip instead as hostname
                        if hits.group(1).endswith(".in-addr.arpa"):
                            host = hits.group(2)
                        else:
                            host = hits.group(1)

                        # if no reverse dns exists, just use ip instead as hostname
                        if hits.group(2) is not None:
                            ip = hits.group(2)
                        else:
                            ip = hits.group(1)

                        if host is not None:
                            # update inventory
                            results.append(dict())
                            results[-1]["name"] = host
                            results[-1]["ip"] = ip
                            ports = []
                        continue

                    host_ports = self.find_port.match(line)
                    if host is not None and host_ports:
                        ports.append(
                            {
                                "port": host_ports.group(1),
                                "protocol": host_ports.group(2),
                                "state": host_ports.group(3),
                                "service": host_ports.group(4),
                            }
                        )
                        continue

                # if any leftovers
                if host and ports:
                    results[-1]["ports"] = ports

            except Exception as e:
                raise AnsibleParserError(f"failed to parse {to_native(path)}: {e} ") from e

        if cache_needs_update:
            self._cache[cache_key] = results

        self._populate(results)



#!/usr/bin/python
#
# Copyright (c) 2019 Yuwei Zhou, <yuwzho@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_iotdevice
version_added: "0.1.2"
short_description: Manage Azure IoT hub device
description:
    - Create, delete an Azure IoT hub device.
options:
    hub:
        description:
            - Name of IoT Hub.
        type: str
        required: true
    hub_policy_name:
        description:
            - Policy name of the IoT Hub which will be used to query from IoT hub.
            - This policy should have 'RegistryWrite, ServiceConnect, DeviceConnect' accesses. You may get 401 error when you lack any of these.
        type: str
        required: true
    hub_policy_key:
        description:
            - Key of the I(hub_policy_name).
        type: str
        required: true
    name:
        description:
            - Name of the IoT hub device identity.
        type: str
        required: true
    state:
        description:
            - State of the IoT hub. Use C(present) to create or update an IoT hub device and C(absent) to delete an IoT hub device.
        type: str
        default: present
        choices:
            - absent
            - present
    auth_method:
        description:
            - The authorization type an entity is to be created with.
        type: str
        choices:
            - sas
            - certificate_authority
            - self_signed
        default: sas
    primary_key:
        description:
            - Explicit self-signed certificate thumbprint to use for primary key.
            - Explicit Shared Private Key to use for primary key.
        type: str
        aliases:
            - primary_thumbprint
    secondary_key:
        description:
            - Explicit self-signed certificate thumbprint to use for secondary key.
            - Explicit Shared Private Key to use for secondary key.
        type: str
        aliases:
            - secondary_thumbprint
    status:
        description:
            - Set device status upon creation.
        type: bool
    edge_enabled:
        description:
            - Flag indicating edge enablement.
            - Not supported in IoT Hub with Basic tier.
        type: bool
    twin_tags:
        description:
            - A section that the solution back end can read from and write to.
            - Tags are not visible to device apps.
            - "The tag can be nested dictionary, '.', '$', '#', ' ' is not allowed in the key."
            - List is not supported.
            - Not supported in IoT Hub with Basic tier.
        type: dict
    desired:
        description:
            - Used along with reported properties to synchronize device configuration or conditions.
            - "The tag can be nested dictionary, '.', '$', '#', ' ' is not allowed in the key."
            - List is not supported.
            - Not supported in IoT Hub with Basic tier.
        type: dict
    device_scope:
        description:
            - The scope of the device. Default value is C(None).
            - The I(device_scpe) shoud start with 'ms-azure-iot-edge://'. Sample as C(ms-azure-iot-edge://{{ edge_device_name }}-{{ generation_id }}).
        type: str
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Yuwei Zhou (@yuwzho)

'''

EXAMPLES = '''
- name: Create simplest Azure IoT Hub device
  azure_rm_iotdevice:
    hub: myHub
    name: Testing
    hub_policy_name: iothubowner
    hub_policy_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    primary_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    secondary_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"

- name: Create Azure IoT Edge device
  azure_rm_iotdevice:
    hub: myHub
    name: Testing
    hub_policy_name: iothubowner
    hub_policy_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    primary_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    secondary_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    edge_enabled: true

- name: Create Azure IoT Hub device with device twin properties and tag
  azure_rm_iotdevice:
    hub: myHub
    name: Testing
    hub_policy_name: iothubowner
    hub_policy_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    primary_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    secondary_key: "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    twin_tags:
      location:
        country: US
        city: Redmond
      sensor: humidity
    desired:
      period: 100
'''

RETURN = '''
device:
    description:
        - IoT Hub device.
    returned: always
    type: dict
    sample: {
        "authentication": {
            "symmetricKey": {
                "primaryKey": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                "secondaryKey": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
            },
            "type": "sas",
            "x509Thumbprint": {
                "primaryThumbprint": null,
                "secondaryThumbprint": null
            }
        },
        "capabilities": {
            "iotEdge": false
        },
        "changed": true,
        "cloudToDeviceMessageCount": 0,
        "connectionState": "Disconnected",
        "connectionStateUpdatedTime": "0001-01-01T00:00:00",
        "deviceId": "Testing",
        "etag": "NzA2NjU2ODc=",
        "failed": false,
        "generationId": "636903014505613307",
        "lastActivityTime": "0001-01-01T00:00:00",
        "modules": [
            {
                "authentication": {
                    "symmetricKey": {
                        "primaryKey": "XXXXXXXXXXXXXXXXXXX",
                        "secondaryKey": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
                    },
                    "type": "sas",
                    "x509Thumbprint": {
                        "primaryThumbprint": null,
                        "secondaryThumbprint": null
                    }
                },
                "cloudToDeviceMessageCount": 0,
                "connectionState": "Disconnected",
                "connectionStateUpdatedTime": "0001-01-01T00:00:00",
                "deviceId": "testdevice",
                "etag": "MjgxOTE5ODE4",
                "generationId": "636903840872788074",
                "lastActivityTime": "0001-01-01T00:00:00",
                "managedBy": null,
                "moduleId": "test"
            }
        ],
        "properties": {
            "desired": {
                "$metadata": {
                    "$lastUpdated": "2019-04-10T05:00:46.2702079Z",
                    "$lastUpdatedVersion": 8,
                    "period": {
                        "$lastUpdated": "2019-04-10T05:00:46.2702079Z",
                        "$lastUpdatedVersion": 8
                    }
                },
                "$version": 1,
                "period": 100
            },
            "reported": {
                "$metadata": {
                    "$lastUpdated": "2019-04-08T06:24:10.5613307Z"
                },
                "$version": 1
            }
        },
        "status": "enabled",
        "statusReason": null,
        "statusUpdatedTime": "0001-01-01T00:00:00",
        "tags": {
            "location": {
                "country": "us",
                "city": "Redmond"
            },
            "sensor": "humidity"
        }
    }
'''  # NOQA

import re
import traceback

from ansible.module_utils.basic import missing_required_lib
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from azure.iot.hub import IoTHubRegistryManager
except ImportError:
    AZURE_IOT_HUB_IMPORT_ERROR = traceback.format_exc()
else:
    AZURE_IOT_HUB_IMPORT_ERROR = None


class AzureRMIoTDevice(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(type='str', required=True),
            hub_policy_name=dict(type='str', required=True),
            hub_policy_key=dict(type='str', no_log=True, required=True),
            hub=dict(type='str', required=True),
            state=dict(type='str', default='present', choices=['present', 'absent']),
            status=dict(type='bool'),
            edge_enabled=dict(type='bool'),
            twin_tags=dict(type='dict'),
            desired=dict(type='dict'),
            auth_method=dict(type='str', choices=['self_signed', 'sas', 'certificate_authority'], default='sas'),
            primary_key=dict(type='str', no_log=True, aliases=['primary_thumbprint']),
            secondary_key=dict(type='str', no_log=True, aliases=['secondary_thumbprint']),
            device_scope=dict(type='str'),
        )

        self.results = dict(
            changed=False,
            id=None
        )

        self.name = None
        self.hub = None
        self.hub_policy_key = None
        self.hub_policy_name = None
        self.state = None
        self.status = None
        self.edge_enabled = None
        self.twin_tags = None
        self.desired = None
        self.auth_method = None
        self.primary_key = None
        self.secondary_key = None
        self.device_scope = None

        self._base_url = None
        self.mgmt_client = None
        super(AzureRMIoTDevice, self).__init__(self.module_arg_spec, supports_check_mode=True)

        if AZURE_IOT_HUB_IMPORT_ERROR:
            self.fail(msg=missing_required_lib('azure-iot-hub', reason=("This module requires the Azure IoT Hub data-plane SDK (azure-iot-hub). "
                                                                        "Installation may fail on some platforms due to uamqp wheel build issues.")),
                      exception=AZURE_IOT_HUB_IMPORT_ERROR)

    def exec_module(self, **kwargs):

        for key in self.module_arg_spec.keys():
            setattr(self, key, kwargs[key])

        self._base_url = '{0}.azure-devices.net'.format(self.hub)

        connect_str = "HostName={0};SharedAccessKeyName={1};SharedAccessKey={2}".format(self._base_url, self.hub_policy_name, self.hub_policy_key)
        self.mgmt_client = IoTHubRegistryManager.from_connection_string(connect_str)

        changed = False
        update_changed = False

        device = self.get_device()

        if self.status is not None and not self.status:
            device['status'] = 'disabled'

        if self.state == 'present':
            if not device:
                changed = True
            else:
                update_changed = True
                if self.edge_enabled is not None and self.edge_enabled != device['capabilities']['iotEdge']:
                    changed = True
                    device['capabilities']['iotEdge'] = self.edge_enabled
                if self.status is not None:
                    status = 'enabled' if self.status else 'disabled'
                    if status != device['status']:
                        changed = True
                        device['status'] = status
                if self.device_scope is not None and self.device_scope != device['device_scope']:
                    changed = True
            if not self.check_mode:
                if changed and update_changed:
                    device = self.update_device(device)
                elif changed:
                    device = self.create_device()

            twin = self.get_twin()

            if twin:
                if not twin.get('tags'):
                    twin['tags'] = dict()
                twin_change = False
                if self.twin_tags and not self.is_equal(self.twin_tags, twin['tags']):
                    twin_change = True
                if self.desired and not self.is_equal(self.desired, twin['properties']['desired']):
                    twin_change = True
                if twin_change and not self.check_mode:
                    twin_dict = dict()
                    twin_dict['tags'] = self.twin_tags
                    twin_dict['properties'] = dict()
                    twin_dict['properties']['desired'] = self.desired

                    twin = self.update_twin(twin_dict)

                changed = changed or twin_change
                device['tags'] = twin.get('tags') or dict()
                device['properties'] = twin.get('properties') or dict()
                device['modules'] = self.list_device_modules()
            elif self.twin_tags or self.desired:
                self.fail("Device twin is not supported in IoT Hub with basic tier.")
        else:
            if not self.check_mode:
                if device:
                    changed = True
                    self.delete_device(device['etag'])
            else:
                changed = True
        self.results = device or dict()
        self.results['changed'] = changed
        return self.results

    def is_equal(self, updated, original):
        changed = False
        if not isinstance(updated, dict):
            self.fail('The Property or Tag should be a dict')
        for key in updated.keys():
            if re.search(r'[.|$|#|\s]', key):
                self.fail("Property or Tag name has invalid characters: '.', '$', '#' or ' '. Got '{0}'".format(key))
            original_value = original.get(key)
            updated_value = updated[key]
            if isinstance(updated_value, dict):
                if not isinstance(original_value, dict):
                    changed = True
                    original[key] = updated_value
                elif not self.is_equal(updated_value, original_value):
                    changed = True
            elif original_value != updated_value:
                changed = True
                original[key] = updated_value
        return not changed

    def update_device(self, device):
        response = None
        try:
            if self.auth_method == 'sas':
                response = self.mgmt_client.update_device_with_sas(self.name, device['etag'],
                                                                   self.primary_key, self.secondary_key, self.status, iot_edge=self.edge_enabled,
                                                                   device_scope=self.device_scope)
            elif self.auth_method == 'self_signed':
                response = self.mgmt_client.update_device_with_certificate_authority(self.name, self.status, iot_edge=self.edge_enabled,
                                                                                     device_scope=self.device_scope)
            elif self.auth_method == 'certificate_authority':
                response = self.mgmt_client.update_device_with_x509(self.name, device['etag'], self.primary_key, self.secondary_key,
                                                                    self.status, iot_edge=self.edge_enabled, device_scope=self.device_scope)

            return self.format_item(response)
        except Exception as exc:
            self.fail('Error when creating or updating IoT Hub device {0}: {1}'.format(self.name, exc.message or str(exc)))

    def create_device(self):
        response = None
        try:
            if self.auth_method == 'sas':
                response = self.mgmt_client.create_device_with_sas(self.name, self.primary_key, self.secondary_key,
                                                                   self.status, iot_edge=self.edge_enabled, device_scope=self.device_scope)
            elif self.auth_method == 'self_signed':
                response = self.mgmt_client.create_device_with_certificate_authority(self.name, self.status,
                                                                                     iot_edge=self.edge_enabled, device_scope=self.device_scope)
            elif self.auth_method == 'certificate_authority':
                response = self.mgmt_client.create_device_with_x509(self.name,
                                                                    self.primary_key, self.secondary_key, self.status,
                                                                    iot_edge=self.edge_enabled, device_scope=self.device_scope)

            return self.format_item(response)
        except Exception as exc:
            self.fail('Error when creating or updating IoT Hub device {0}: {1}'.format(self.name, exc.message or str(exc)))

    def delete_device(self, etag):
        try:
            response = self.mgmt_client.delete_device(self.name, etag=etag)
            return response

        except Exception as exc:
            self.fail('Error when deleting IoT Hub device {0}: {1}'.format(self.name, exc.message or str(exc)))

    def get_device(self):
        try:
            response = self.mgmt_client.get_device(self.name)

            response = self.format_item(response)
            return response
        except Exception as exc:
            self.log('Error when getting IoT Hub device {0}: {1}'.format(self.name, exc))

    def get_twin(self):
        try:
            response = self.mgmt_client.get_twin(self.name)
            return self.format_twin(response)
        except Exception as exc:
            self.fail('Error when getting IoT Hub device {0} twin: {1}'.format(self.name, exc.message or str(exc)))

    def update_twin(self, twin):
        try:
            response = self.mgmt_client.update_twin(self.name, twin)

            return self.format_twin(response)
        except Exception as exc:
            self.fail('Error when creating or updating IoT Hub device twin {0}: {1}'.format(self.name, exc.message or str(exc)))

    def list_device_modules(self):
        try:
            response = None
            response = self.mgmt_client.get_modules(self.name)

            response = [self.format_item(item) for item in response]
            return response

        except Exception as exc:
            if hasattr(exc, 'message'):
                pass
            else:
                self.fail('Error when listing IoT Hub devices in {0}: {1}'.format(self.hub, exc))

    def format_twin(self, item):
        if not item:
            return None
        format_twin = dict(
            device_id=item.device_id,
            module_id=item.module_id,
            tags=item.tags,
            properties=dict(),
            etag=item.etag,
            version=item.version,
            device_etag=item.device_etag,
            status=item.status,
            cloud_to_device_message_count=item.cloud_to_device_message_count,
            authentication_type=item.authentication_type,
        )
        if item.properties is not None:
            format_twin['properties']['desired'] = item.properties.desired
            format_twin['properties']['reported'] = item.properties.reported

        return format_twin

    def format_item(self, item):
        if not item:
            return None
        format_item = dict(
            authentication=dict(),
            capabilities=dict(),
            cloudToDeviceMessageCount=item.cloud_to_device_message_count,
            connectionState=item.connection_state,
            connectionStateUpdatedTime=item.connection_state_updated_time,
            deviceId=item.device_id,
            etag=item.etag,
            generationId=item.generation_id,
            lastActivityTime=item.last_activity_time,
            device_scope=item.device_scope if hasattr(item, 'device_scope') else None
        )
        if hasattr(item, 'status_updated_time'):
            format_item['statusUpdatedTime'] = item.status_updated_time
        if hasattr(item, 'status_reason'):
            format_item['status_reason'] = item.status_reason
        if hasattr(item, 'status'):
            format_item['status'] = item.status
        if hasattr(item, 'modules'):
            format_item['modules'] = item.modules
        if item.authentication:
            format_item['authentication']['symmetricKey'] = dict()
            format_item['authentication']['symmetricKey']['primaryKey'] = item.authentication.symmetric_key.primary_key
            format_item['authentication']['symmetricKey']['secondaryKey'] = item.authentication.symmetric_key.secondary_key

            format_item['authentication']['type'] = item.authentication.type
            format_item['authentication']["x509Thumbprint"] = dict()
            format_item['authentication']["x509Thumbprint"]["primaryThumbprint"] = item.authentication.x509_thumbprint.primary_thumbprint
            format_item['authentication']["x509Thumbprint"]['secondaryThumbprint'] = item.authentication.x509_thumbprint.secondary_thumbprint
        format_item['capabilities'] = dict()
        if hasattr(item, 'capabilities') and item.capabilities is not None:
            format_item['capabilities']["iotEdge"] = item.capabilities.iot_edge

        return format_item


def main():
    AzureRMIoTDevice()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_camera_roles
short_description: Resource module for organizations _camera _roles
description:
  - Manage operations create, update and delete of the resource organizations _camera
    _roles.
  - Creates new role for this organization.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  appliedOnDevices:
    description: Device tag on which this specified permission is applied.
    elements: dict
    suboptions:
      id:
        description: Device id.
        type: str
      inNetworksWithId:
        description: Network id scope.
        type: str
      inNetworksWithTag:
        description: Network tag scope.
        type: str
      permissionScopeId:
        description: Permission scope id.
        type: str
      tag:
        description: Device tag.
        type: str
    type: list
  appliedOnNetworks:
    description: Network tag on which this specified permission is applied.
    elements: dict
    suboptions:
      id:
        description: Network id.
        type: str
      permissionScopeId:
        description: Permission scope id.
        type: str
      tag:
        description: Network tag.
        type: str
    type: list
  appliedOrgWide:
    description: Permissions to be applied org wide.
    elements: dict
    suboptions:
      permissionScopeId:
        description: Permission scope id.
        type: str
    type: list
  name:
    description: The name of the new role. Must be unique. This parameter is required.
    type: str
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for camera createOrganizationCameraRole
    description: Complete reference of the createOrganizationCameraRole API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-camera-role
notes:
  - SDK Method used are
    camera.Camera.create_organization_camera_role,
  - Paths used are
    post /organizations/{organizationId}/camera/roles,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_camera_roles:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    appliedOnDevices:
      - id: ''
        permissionScopeId: '1'
        tag: reception-desk
    appliedOnNetworks:
      - id: ''
        permissionScopeId: '2'
        tag: building-a
    appliedOrgWide:
      - id: ''
        permissionScopeId: '2'
        tag: building-a
    name: Security_Guard
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "name": "string",
      "appliedOnDevices": [
        {
          "tag": "string",
          "id": "string",
          "permissionScopeId": "string",
          "permissionScope": "string",
          "permissionLevel": "string"
        }
      ],
      "appliedOnNetworks": [
        {
          "tag": "string",
          "id": "string",
          "permissionScopeId": "string",
          "permissionScope": "string",
          "permissionLevel": "string"
        }
      ],
      "appliedOrgWide": [
        {
          "tag": "string",
          "permissionScopeId": "string",
          "permissionScope": "string",
          "permissionLevel": "string"
        }
      ]
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: autoscaling_policy
short_description: Create or delete AWS scaling policies for Autoscaling groups
version_added: 1.0.0
description:
  - Can create or delete scaling policies for autoscaling groups.
  - Referenced autoscaling groups must already exist.
  - Prior to release 5.0.0 this module was called C(community.aws.ec2_scaling_policy).
    The usage did not change.
author:
  - Zacharie Eakin (@zeekin)
  - Will Thames (@willthames)
options:
  state:
    type: str
    description:
      - Register or deregister the policy.
    choices: ['present', 'absent']
    default: 'present'
  name:
    type: str
    description:
      - Unique name for the scaling policy.
    required: true
  asg_name:
    type: str
    description:
      - Name of the associated autoscaling group.
      - Required if I(state) is C(present).
  adjustment_type:
    type: str
    description:
      - The type of change in capacity of the autoscaling group.
      - Required if I(state) is C(present).
    choices:
      - ChangeInCapacity
      - ExactCapacity
      - PercentChangeInCapacity
  scaling_adjustment:
    type: int
    description:
      - The amount by which the autoscaling group is adjusted by the policy.
      - A negative number has the effect of scaling down the ASG.
      - Units are numbers of instances for C(ExactCapacity) or C(ChangeInCapacity) or percent
        of existing instances for C(PercentChangeInCapacity).
      - Required when I(policy_type) is C(SimpleScaling).
  min_adjustment_step:
    type: int
    description:
      - Minimum amount of adjustment when policy is triggered.
      - Only used when I(adjustment_type) is C(PercentChangeInCapacity).
  cooldown:
    type: int
    description:
      - The minimum period of time (in seconds) between which autoscaling actions can take place.
      - Only used when I(policy_type) is C(SimpleScaling).
  policy_type:
    type: str
    description:
      - Auto scaling adjustment policy.
    choices:
      - StepScaling
      - SimpleScaling
      - TargetTrackingScaling
    default: SimpleScaling
  metric_aggregation:
    type: str
    description:
      - The aggregation type for the CloudWatch metrics.
      - Only used when I(policy_type) is not C(SimpleScaling).
    choices:
      - Minimum
      - Maximum
      - Average
    default: Average
  step_adjustments:
    type: list
    description:
      - List of dicts containing I(lower_bound), I(upper_bound) and I(scaling_adjustment).
      - Intervals must not overlap or have a gap between them.
      - At most, one item can have an undefined I(lower_bound).
        If any item has a negative lower_bound, then there must be a step adjustment with an undefined I(lower_bound).
      - At most, one item can have an undefined I(upper_bound).
        If any item has a positive upper_bound, then there must be a step adjustment with an undefined I(upper_bound).
      - The bounds are the amount over the alarm threshold at which the adjustment will trigger.
        This means that for an alarm threshold of 50, triggering at 75 requires a lower bound of 25.
        See U(http://docs.aws.amazon.com/AutoScaling/latest/APIReference/API_StepAdjustment.html).
    elements: dict
    suboptions:
      lower_bound:
        type: int
        description:
          - The lower bound for the difference between the alarm threshold and
            the CloudWatch metric.
      upper_bound:
        type: int
        description:
          - The upper bound for the difference between the alarm threshold and
            the CloudWatch metric.
      scaling_adjustment:
        type: int
        description:
          - The amount by which to scale.
        required: true
  target_tracking_config:
    type: dict
    description:
      - Allows you to specify a I(target_tracking_config) for autoscaling policies in AWS.
      - I(target_tracking_config) can accept nested dicts for I(customized_metric_spec) or I(predefined_metric_spec).
        Each specification aligns with their boto3 equivalent.
      - Required when I(TargetTrackingScaling) policy is specified.
    version_added: 4.1.0
    suboptions:
      customized_metric_spec:
        type: dict
        description:
          - Specify a dict will be passed in as a call for C(TargetTrackingConfiguration).
        suboptions:
            metric_name:
              type: str
              description:
                - The name of the metric.
              required: true
            namespace:
              type: str
              description:
                - The namespace of the metric.
              required: true
            statistic:
              type: str
              description:
                - The statistic of the metric.
              required: true
              choices:
                - Average
                - Minimum
                - Maximum
                - SampleCount
                - Sum
            dimensions:
              type: list
              description:
                - The dimensions of the metric. The element of the list should be a dict.
              elements: dict
            unit:
              type: str
              description:
                - The unit of the metric. Reference AmazonCloudWatch API for valid Units.
      predefined_metric_spec:
        type: dict
        description:
          - Specify a dict will be passed in as a call for I(TargetTrackingConfiguration).
        suboptions:
          predefined_metric_type:
            type: str
            required: true
            description:
              - Required if C(predefined_metric_spec) is used.
            choices:
              - ASGAverageCPUUtilization
              - ASGAverageNetworkIn
              - ASGAverageNetworkOut
              - ALBRequestCountPerTarget
          resource_label:
            type: str
            description:
              - Uniquely identifies a specific ALB target group from which to determine the average request count served by your Auto Scaling group.
              - You can't specify a resource label unless the target group is attached to the Auto Scaling group.
      target_value:
        type: float
        description:
          - Specify a float number for target utilization.
          - Required when I(target_tracking_config) is specified.
        required: true
      disable_scalein:
        type: bool
        description:
          - Indicate whether scaling in by the target tracking scaling policy is disabled.
  estimated_instance_warmup:
    type: int
    description:
      - The estimated time, in seconds, until a newly launched instance can contribute to the CloudWatch metrics.
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: Simple Scale Down policy
  community.aws.autoscaling_policy:
    state: present
    region: US-XXX
    name: "scaledown-policy"
    adjustment_type: "ChangeInCapacity"
    asg_name: "application-asg"
    scaling_adjustment: -1
    min_adjustment_step: 1
    cooldown: 300

# For an alarm with a breach threshold of 20, the
# following creates a stepped policy:
# From 20-40 (0-20 above threshold), increase by 50% of existing capacity
# From 41-infinity, increase by 100% of existing capacity
- community.aws.autoscaling_policy:
    state: present
    region: US-XXX
    name: "step-scale-up-policy"
    policy_type: StepScaling
    metric_aggregation: Maximum
    step_adjustments:
      - upper_bound: 20
        scaling_adjustment: 50
      - lower_bound: 20
        scaling_adjustment: 100
    adjustment_type: "PercentChangeInCapacity"
    asg_name: "application-asg"

- name: create TargetTracking predefined policy
  community.aws.autoscaling_policy:
    name: "predefined-policy-1"
    policy_type: TargetTrackingScaling
    target_tracking_config:
      predefined_metric_spec:
        predefined_metric_type: ASGAverageCPUUtilization
      target_value: 98.0
    asg_name: "asg-test-1"
  register: result

- name: create TargetTracking predefined policy with resource_label
  community.aws.autoscaling_policy:
    name: "predefined-policy-1"
    policy_type: TargetTrackingScaling
    target_tracking_config:
      predefined_metric_spec:
        predefined_metric_type: ALBRequestCountPerTarget
        resource_label: app/my-alb/778d41231d141a0f/targetgroup/my-alb-target-group/942f017f100becff
      target_value: 98.0
    asg_name: "asg-test-1"
  register: result

- name: create TargetTrackingScaling custom policy
  community.aws.autoscaling_policy:
    name: "custom-policy-1"
    policy_type: TargetTrackingScaling
    target_tracking_config:
      customized_metric_spec:
        metric_name: metric_1
        namespace: namespace_1
        statistic: Minimum
        unit: Gigabits
        dimensions: [{'Name': 'dimension1', 'Value': 'value1'}]
      disable_scalein: true
      target_value: 98.0
    asg_name: asg-test-1
  register: result
"""

RETURN = r"""
adjustment_type:
  description: Scaling policy adjustment type.
  returned: always
  type: str
  sample: PercentChangeInCapacity
alarms:
  description: Cloudwatch alarms related to the policy.
  returned: always
  type: complex
  contains:
    alarm_name:
      description: Name of the Cloudwatch alarm.
      returned: always
      type: str
      sample: cpu-very-high
    alarm_arn:
      description: ARN of the Cloudwatch alarm.
      returned: always
      type: str
      sample: arn:aws:cloudwatch:us-east-2:1234567890:alarm:cpu-very-high
arn:
  description: ARN of the scaling policy. Provided for backward compatibility, value is the same as I(policy_arn).
  returned: always
  type: str
  sample: arn:aws:autoscaling:us-east-2:123456789012:scalingPolicy:59e37526-bd27-42cf-adca-5cd3d90bc3b9:autoScalingGroupName/app-asg:policyName/app-policy
as_name:
  description: Auto Scaling Group name. Provided for backward compatibility, value is the same as I(auto_scaling_group_name).
  returned: always
  type: str
  sample: app-asg
auto_scaling_group_name:
  description: Name of Auto Scaling Group.
  returned: always
  type: str
  sample: app-asg
metric_aggregation_type:
  description: Method used to aggregate metrics.
  returned: when I(policy_type) is C(StepScaling)
  type: str
  sample: Maximum
name:
  description: Name of the scaling policy. Provided for backward compatibility, value is the same as I(policy_name).
  returned: always
  type: str
  sample: app-policy
policy_arn:
  description: ARN of scaling policy.
  returned: always
  type: str
  sample: arn:aws:autoscaling:us-east-2:123456789012:scalingPolicy:59e37526-bd27-42cf-adca-5cd3d90bc3b9:autoScalingGroupName/app-asg:policyName/app-policy
policy_name:
  description: Name of scaling policy.
  returned: always
  type: str
  sample: app-policy
policy_type:
  description: Type of auto scaling policy.
  returned: always
  type: str
  sample: StepScaling
scaling_adjustment:
  description: Adjustment to make when alarm is triggered.
  returned: When I(policy_type) is C(SimpleScaling)
  type: int
  sample: 1
step_adjustments:
  description: List of step adjustments.
  returned: always
  type: complex
  contains:
    metric_interval_lower_bound:
      description: Lower bound for metric interval.
      returned: if step has a lower bound
      type: float
      sample: 20.0
    metric_interval_upper_bound:
      description: Upper bound for metric interval.
      returned: if step has an upper bound
      type: float
      sample: 40.0
    scaling_adjustment:
      description: Adjustment to make if this step is reached.
      returned: always
      type: int
      sample: 50
"""

try:
    import botocore
except ImportError:
    pass  # caught by imported AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def build_target_specification(target_tracking_config):
    # Initialize an empty dict() for building TargetTrackingConfiguration policies,
    # which will be returned
    targetTrackingConfig = dict()

    if target_tracking_config.get("target_value"):
        targetTrackingConfig["TargetValue"] = target_tracking_config["target_value"]

    if target_tracking_config.get("disable_scalein"):
        targetTrackingConfig["DisableScaleIn"] = target_tracking_config["disable_scalein"]
    else:
        # Accounting for boto3 response
        targetTrackingConfig["DisableScaleIn"] = False

    if target_tracking_config["predefined_metric_spec"] is not None:
        # Build spec for predefined_metric_spec
        targetTrackingConfig["PredefinedMetricSpecification"] = dict()
        if target_tracking_config["predefined_metric_spec"].get("predefined_metric_type"):
            targetTrackingConfig["PredefinedMetricSpecification"]["PredefinedMetricType"] = target_tracking_config[
                "predefined_metric_spec"
            ]["predefined_metric_type"]

        if target_tracking_config["predefined_metric_spec"].get("resource_label"):
            targetTrackingConfig["PredefinedMetricSpecification"]["ResourceLabel"] = target_tracking_config[
                "predefined_metric_spec"
            ]["resource_label"]

    elif target_tracking_config["customized_metric_spec"] is not None:
        # Build spec for customized_metric_spec
        targetTrackingConfig["CustomizedMetricSpecification"] = dict()
        if target_tracking_config["customized_metric_spec"].get("metric_name"):
            targetTrackingConfig["CustomizedMetricSpecification"]["MetricName"] = target_tracking_config[
                "customized_metric_spec"
            ]["metric_name"]

        if target_tracking_config["customized_metric_spec"].get("namespace"):
            targetTrackingConfig["CustomizedMetricSpecification"]["Namespace"] = target_tracking_config[
                "customized_metric_spec"
            ]["namespace"]

        if target_tracking_config["customized_metric_spec"].get("dimensions"):
            targetTrackingConfig["CustomizedMetricSpecification"]["Dimensions"] = target_tracking_config[
                "customized_metric_spec"
            ]["dimensions"]

        if target_tracking_config["customized_metric_spec"].get("statistic"):
            targetTrackingConfig["CustomizedMetricSpecification"]["Statistic"] = target_tracking_config[
                "customized_metric_spec"
            ]["statistic"]

        if target_tracking_config["customized_metric_spec"].get("unit"):
            targetTrackingConfig["CustomizedMetricSpecification"]["Unit"] = target_tracking_config[
                "customized_metric_spec"
            ]["unit"]

    return targetTrackingConfig


def create_scaling_policy(connection, module):
    changed = False
    asg_name = module.params["asg_name"]
    policy_type = module.params["policy_type"]
    policy_name = module.params["name"]

    if policy_type == "TargetTrackingScaling":
        params = dict(PolicyName=policy_name, PolicyType=policy_type, AutoScalingGroupName=asg_name)
    else:
        params = dict(
            PolicyName=policy_name,
            PolicyType=policy_type,
            AutoScalingGroupName=asg_name,
            AdjustmentType=module.params["adjustment_type"],
        )

    # min_adjustment_step attribute is only relevant if the adjustment_type
    # is set to percentage change in capacity, so it is a special case
    if module.params["adjustment_type"] == "PercentChangeInCapacity":
        if module.params["min_adjustment_step"]:
            params["MinAdjustmentMagnitude"] = module.params["min_adjustment_step"]

    if policy_type == "SimpleScaling":
        # can't use required_if because it doesn't allow multiple criteria -
        # it's only required if policy is SimpleScaling and state is present
        if not module.params["scaling_adjustment"]:
            module.fail_json(
                msg="scaling_adjustment is required when policy_type is SimpleScaling and state is present"
            )
        params["ScalingAdjustment"] = module.params["scaling_adjustment"]
        if module.params["cooldown"]:
            params["Cooldown"] = module.params["cooldown"]

    elif policy_type == "StepScaling":
        if not module.params["step_adjustments"]:
            module.fail_json(msg="step_adjustments is required when policy_type is StepScaling and state is present")
        params["StepAdjustments"] = []
        for step_adjustment in module.params["step_adjustments"]:
            step_adjust_params = dict(ScalingAdjustment=step_adjustment["scaling_adjustment"])
            if step_adjustment.get("lower_bound"):
                step_adjust_params["MetricIntervalLowerBound"] = step_adjustment["lower_bound"]
            if step_adjustment.get("upper_bound"):
                step_adjust_params["MetricIntervalUpperBound"] = step_adjustment["upper_bound"]
            params["StepAdjustments"].append(step_adjust_params)
        if module.params["metric_aggregation"]:
            params["MetricAggregationType"] = module.params["metric_aggregation"]
        if module.params["estimated_instance_warmup"]:
            params["EstimatedInstanceWarmup"] = module.params["estimated_instance_warmup"]

    elif policy_type == "TargetTrackingScaling":
        if not module.params["target_tracking_config"]:
            module.fail_json(
                msg="target_tracking_config is required when policy_type is TargetTrackingScaling and state is present"
            )
        else:
            params["TargetTrackingConfiguration"] = build_target_specification(
                module.params.get("target_tracking_config")
            )
        if module.params["estimated_instance_warmup"]:
            params["EstimatedInstanceWarmup"] = module.params["estimated_instance_warmup"]

    # Ensure idempotency with policies
    try:
        policies = connection.describe_policies(
            aws_retry=True, AutoScalingGroupName=asg_name, PolicyNames=[policy_name]
        )["ScalingPolicies"]
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg=f"Failed to obtain autoscaling policy {policy_name}")

    before = after = {}
    if not policies:
        changed = True
    else:
        policy = policies[0]
        for key, old_value in params.items():
            if old_value != policy.get(key):
                changed = True
                before[key] = old_value
                after[key] = policy.get(key)

    if changed:
        try:
            connection.put_scaling_policy(aws_retry=True, **params)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            module.fail_json_aws(e, msg="Failed to create autoscaling policy")

        try:
            policies = connection.describe_policies(
                aws_retry=True, AutoScalingGroupName=asg_name, PolicyNames=[policy_name]
            )["ScalingPolicies"]
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            module.fail_json_aws(e, msg=f"Failed to obtain autoscaling policy {policy_name}")

    policy = camel_dict_to_snake_dict(policies[0])
    # Backward compatible return values
    policy["arn"] = policy["policy_arn"]
    policy["as_name"] = policy["auto_scaling_group_name"]
    policy["name"] = policy["policy_name"]

    if before and after:
        module.exit_json(changed=changed, diff=dict(before=before, after=after), **policy)
    else:
        module.exit_json(changed=changed, **policy)


def delete_scaling_policy(connection, module):
    policy_name = module.params.get("name")

    try:
        policy = connection.describe_policies(aws_retry=True, PolicyNames=[policy_name])
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg=f"Failed to obtain autoscaling policy {policy_name}")

    if policy["ScalingPolicies"]:
        try:
            connection.delete_policy(
                aws_retry=True,
                AutoScalingGroupName=policy["ScalingPolicies"][0]["AutoScalingGroupName"],
                PolicyName=policy_name,
            )
            module.exit_json(changed=True)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            module.fail_json_aws(e, msg="Failed to delete autoscaling policy")

    module.exit_json(changed=False)


def main():
    step_adjustment_spec = dict(
        lower_bound=dict(type="int"), upper_bound=dict(type="int"), scaling_adjustment=dict(type="int", required=True)
    )

    predefined_metric_spec = dict(
        predefined_metric_type=dict(
            type="str",
            choices=[
                "ASGAverageCPUUtilization",
                "ASGAverageNetworkIn",
                "ASGAverageNetworkOut",
                "ALBRequestCountPerTarget",
            ],
            required=True,
        ),
        resource_label=dict(type="str"),
    )
    customized_metric_spec = dict(
        metric_name=dict(type="str", required=True),
        namespace=dict(type="str", required=True),
        statistic=dict(type="str", required=True, choices=["Average", "Minimum", "Maximum", "SampleCount", "Sum"]),
        dimensions=dict(type="list", elements="dict"),
        unit=dict(type="str"),
    )

    target_tracking_spec = dict(
        disable_scalein=dict(type="bool"),
        target_value=dict(type="float", required=True),
        predefined_metric_spec=dict(type="dict", options=predefined_metric_spec),
        customized_metric_spec=dict(type="dict", options=customized_metric_spec),
    )

    argument_spec = dict(
        name=dict(required=True),
        adjustment_type=dict(choices=["ChangeInCapacity", "ExactCapacity", "PercentChangeInCapacity"]),
        asg_name=dict(),
        scaling_adjustment=dict(type="int"),
        min_adjustment_step=dict(type="int"),
        cooldown=dict(type="int"),
        state=dict(default="present", choices=["present", "absent"]),
        metric_aggregation=dict(default="Average", choices=["Minimum", "Maximum", "Average"]),
        policy_type=dict(default="SimpleScaling", choices=["SimpleScaling", "StepScaling", "TargetTrackingScaling"]),
        target_tracking_config=dict(type="dict", options=target_tracking_spec),
        step_adjustments=dict(type="list", options=step_adjustment_spec, elements="dict"),
        estimated_instance_warmup=dict(type="int"),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, required_if=[["state", "present", ["asg_name"]]])

    connection = module.client("autoscaling", retry_decorator=AWSRetry.jittered_backoff())
    state = module.params.get("state")

    if state == "present":
        create_scaling_policy(connection, module)
    elif state == "absent":
        delete_scaling_policy(connection, module)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2020, Lukas Bestle <project-ansible@lukasbestle.com>
# Copyright (c) 2017, Michael Heap <m@michaelheap.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: mas
short_description: Manage Mac App Store applications with mas-cli
description:
  - Installs, uninstalls and updates macOS applications from the Mac App Store using the C(mas-cli).
version_added: '0.2.0'
author:
  - Michael Heap (@mheap)
  - Lukas Bestle (@lukasbestle)
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  id:
    description:
      - The Mac App Store identifier of the app(s) you want to manage.
      - This can be found by running C(mas search APP_NAME) on your machine.
    type: list
    elements: int
  state:
    description:
      - Desired state of the app installation.
      - The V(absent) value requires root permissions, also see the examples.
    type: str
    choices:
      - absent
      - latest
      - present
    default: present
  upgrade_all:
    description:
      - Upgrade all installed Mac App Store apps.
    type: bool
    default: false
    aliases: ["upgrade"]
requirements:
  - macOS 10.11 or higher.
  - "mas-cli (U(https://github.com/mas-cli/mas)) 1.5.0+ available as C(mas) in the bin path"
  - The Apple ID to use already needs to be signed in to the Mac App Store (check with C(mas account)).
  - The feature of "checking if user is signed in" is disabled for anyone using macOS 12.0+.
  - Users need to sign in to the Mac App Store GUI beforehand for anyone using macOS 12.0+ due to U(https://github.com/mas-cli/mas/issues/417).
"""

EXAMPLES = r"""
- name: Install Keynote
  community.general.mas:
    id: 409183694
    state: present

- name: Install Divvy with command mas installed in /usr/local/bin
  community.general.mas:
    id: 413857545
    state: present
  environment:
    PATH: /usr/local/bin:{{ ansible_facts.env.PATH }}

- name: Install a list of apps
  community.general.mas:
    id:
      - 409183694 # Keynote
      - 413857545 # Divvy
    state: present

- name: Ensure the latest Keynote version is installed
  community.general.mas:
    id: 409183694
    state: latest

- name: Upgrade all installed Mac App Store apps
  community.general.mas:
    upgrade_all: true

- name: Install specific apps and also upgrade all others
  community.general.mas:
    id:
      - 409183694 # Keynote
      - 413857545 # Divvy
    state: present
    upgrade_all: true

- name: Uninstall Divvy
  community.general.mas:
    id: 413857545
    state: absent
  become: true # Uninstallation requires root permissions
"""

RETURN = r""" # """

import os
import platform

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._version import LooseVersion

NOT_WORKING_MAC_VERSION_MAS_ACCOUNT = "12.0"


class Mas:
    def __init__(self, module):
        self.module = module

        # Initialize data properties
        self.mas_path = self.module.get_bin_path("mas")
        self._checked_signin = False
        self._mac_version = platform.mac_ver()[0] or "0.0"
        self._installed = None  # Populated only if needed
        self._outdated = None  # Populated only if needed
        self.count_install = 0
        self.count_upgrade = 0
        self.count_uninstall = 0
        self.result = {"changed": False}

        self.check_mas_tool()

    def app_command(self, command, id):
        """Runs a `mas` command on a given app; command can be 'install', 'upgrade' or 'uninstall'"""

        if not self.module.check_mode:
            if command != "uninstall":
                self.check_signin()

            rc, out, err = self.run([command, str(id)])
            if rc != 0:
                self.module.fail_json(msg=f"Error running command '{command}' on app '{id}': {out.rstrip()}")

        # No error or dry run
        self.__dict__[f"count_{command}"] += 1

    def check_mas_tool(self):
        """Verifies that the `mas` tool is available in a recent version"""

        # Is the `mas` tool available at all?
        if not self.mas_path:
            self.module.fail_json(msg="Required `mas` tool is not installed")

        # Is the version recent enough?
        rc, out, err = self.run(["version"])
        if rc != 0 or not out.strip() or LooseVersion(out.strip()) < LooseVersion("1.5.0"):
            self.module.fail_json(msg=f"`mas` tool in version 1.5.0+ needed, got {out.strip()}")

    def check_signin(self):
        """Verifies that the user is signed in to the Mac App Store"""
        # Only check this once per execution
        if self._checked_signin:
            return
        if LooseVersion(self._mac_version) >= LooseVersion(NOT_WORKING_MAC_VERSION_MAS_ACCOUNT):
            # Checking if user is signed-in is disabled due to https://github.com/mas-cli/mas/issues/417
            self.module.log("WARNING: You must be signed in via the Mac App Store GUI beforehand else error will occur")
        else:
            rc, out, err = self.run(["account"])
            if out.split("\n", 1)[0].rstrip() == "Not signed in":
                self.module.fail_json(msg="You must be signed in to the Mac App Store")

        self._checked_signin = True

    def exit(self):
        """Exit with the data we have collected over time"""

        msgs = []
        if self.count_install > 0:
            msgs.append(f"Installed {self.count_install} app(s)")
        if self.count_upgrade > 0:
            msgs.append(f"Upgraded {self.count_upgrade} app(s)")
        if self.count_uninstall > 0:
            msgs.append(f"Uninstalled {self.count_uninstall} app(s)")

        if msgs:
            self.result["changed"] = True
            self.result["msg"] = ", ".join(msgs)

        self.module.exit_json(**self.result)

    def get_current_state(self, command):
        """Returns the list of all app IDs; command can either be 'list' or 'outdated'"""

        rc, raw_apps, err = self.run([command])
        rows = raw_apps.split("\n")
        if rows[0] == "No installed apps found":
            rows = []
        apps = []
        for r in rows:
            # mas 2.3.0 and older:
            # 123456789  App Name         (version)
            # 4567890    App Name Longer  (version)
            #
            # mas 3.0.0 and newer:
            # 123456789  App Name         (version)
            #   4567890  App Name Longer  (version)
            r = r.strip().split(" ", 1)
            if len(r) == 2:
                apps.append(int(r[0]))

        return apps

    def installed(self):
        """Returns the list of installed apps"""

        # Populate cache if not already done
        if self._installed is None:
            self._installed = self.get_current_state("list")

        return self._installed

    def is_installed(self, id):
        """Checks whether the given app is installed"""

        return int(id) in self.installed()

    def is_outdated(self, id):
        """Checks whether the given app is installed, but outdated"""

        return int(id) in self.outdated()

    def outdated(self):
        """Returns the list of installed, but outdated apps"""

        # Populate cache if not already done
        if self._outdated is None:
            self._outdated = self.get_current_state("outdated")

        return self._outdated

    def run(self, cmd):
        """Runs a command of the `mas` tool"""

        cmd.insert(0, self.mas_path)
        return self.module.run_command(cmd, False)

    def upgrade_all(self):
        """Upgrades all installed apps and sets the correct result data"""

        outdated = self.outdated()

        if not self.module.check_mode:
            self.check_signin()

            rc, out, err = self.run(["upgrade"])
            if rc != 0:
                self.module.fail_json(msg=f"Could not upgrade all apps: {out.rstrip()}")

        self.count_upgrade += len(outdated)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            id=dict(type="list", elements="int"),
            state=dict(type="str", default="present", choices=["absent", "latest", "present"]),
            upgrade_all=dict(type="bool", default=False, aliases=["upgrade"]),
        ),
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}
    mas = Mas(module)

    if module.params["id"]:
        apps = module.params["id"]
    else:
        apps = []

    state = module.params["state"]
    upgrade = module.params["upgrade_all"]

    # Run operations on the given app IDs
    for app in sorted(set(apps)):
        if state == "present":
            if not mas.is_installed(app):
                mas.app_command("install", app)

        elif state == "absent":
            if mas.is_installed(app):
                # Ensure we are root
                if os.getuid() != 0:
                    module.fail_json(msg="Uninstalling apps requires root permissions ('become: true')")

                mas.app_command("uninstall", app)

        elif state == "latest":
            if not mas.is_installed(app):
                mas.app_command("install", app)
            elif mas.is_outdated(app):
                mas.app_command("upgrade", app)

    # Upgrade all apps if requested
    mas._outdated = None  # Clear cache
    if upgrade and mas.outdated():
        mas.upgrade_all()

    # Exit with the collected data
    mas.exit()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: acl
short_description: Set and retrieve file ACL information.
description:
- Set and retrieve file ACL information.
version_added: "1.0.0"
options:
  path:
    description:
    - The full path of the file or object.
    type: path
    required: true
    aliases: [ name ]
  state:
    description:
    - Define whether the ACL should be present or not.
    - The V(query) state gets the current ACL without changing it, for use in C(register) operations.
    choices: [ absent, present, query ]
    default: query
    type: str
  follow:
    description:
    - Whether to follow symlinks on the path if a symlink is encountered.
    type: bool
    default: true
  default:
    description:
    - If O(path) is a directory, setting this to V(true) will make it the default ACL for entities created inside the directory.
    - Setting O(default=true) causes an error if O(path) is a file.
    type: bool
    default: false
  entity:
    description:
    - The actual user or group that the ACL applies to when matching entity types user or group are selected.
    type: str
    default: ""
  etype:
    description:
    - The entity type of the ACL to apply, see C(setfacl) documentation for more info.
    choices: [ group, mask, other, user ]
    type: str
  permissions:
    description:
    - The permissions to apply/remove can be any combination of C(r), C(w), C(x)
      (read, write and execute respectively), and C(X) (execute permission if the file is a directory or already has execute permission for some user)
    type: str
  entry:
    description:
    - DEPRECATED.
    - The ACL to set or remove.
    - This must always be quoted in the form of C(<etype>:<qualifier>:<perms>).
    - The qualifier may be empty for some types, but the type and perms are always required.
    - C(-) can be used as placeholder when you do not care about permissions.
    - This is now superseded by entity, type and permissions fields.
    type: str
  recursive:
    description:
    - Recursively sets the specified ACL.
    - Incompatible with O(state=query).
    - Alias O(recurse) added in version 1.3.0.
    type: bool
    default: false
    aliases: [ recurse ]
  use_nfsv4_acls:
    description:
    - Use NFSv4 ACLs instead of POSIX ACLs.
    - This feature uses C(nfs4_setfacl) and C(nfs4_getfacl). The behavior depends on those implementation.
      And currently it only supports C(A) in ACE, so C(D) must be replaced with the appropriate C(A).
    - Permission is set as optimised ACLs by the system. You can check the actual ACLs that has been set using the return value.
    - More info C(man nfs4_setfacl)
    type: bool
    default: false
  recalculate_mask:
    description:
    - Select if and when to recalculate the effective right masks of the files.
    - See C(setfacl) documentation for more info.
    - Incompatible with O(state=query).
    choices: [ default, mask, no_mask ]
    default: default
    type: str
author:
- Brian Coca (@bcoca)
- Jérémie Astori (@astorije)
notes:
- The M(ansible.posix.acl) module requires that ACLs are enabled on the target filesystem and that the C(setfacl) and C(getfacl) binaries are installed.
- As of Ansible 2.0, this module only supports Linux distributions.
- As of Ansible 2.3, the O(name) option has been changed to O(path) as default, but O(name) still works as well.
'''

EXAMPLES = r'''
- name: Grant user Joe read access to a file
  ansible.posix.acl:
    path: /etc/foo.conf
    entity: joe
    etype: user
    permissions: r
    state: present

- name: Removes the ACL for Joe on a specific file
  ansible.posix.acl:
    path: /etc/foo.conf
    entity: joe
    etype: user
    state: absent

- name: Sets default ACL for joe on /etc/foo.d/
  ansible.posix.acl:
    path: /etc/foo.d/
    entity: joe
    etype: user
    permissions: rw
    default: true
    state: present

- name: Same as previous but using entry shorthand
  ansible.posix.acl:
    path: /etc/foo.d/
    entry: default:user:joe:rw-
    state: present

- name: Obtain the ACL for a specific file
  ansible.posix.acl:
    path: /etc/foo.conf
  register: acl_info
'''

RETURN = r'''
acl:
    description: Current ACL on provided path (after changes, if any)
    returned: success
    type: list
    sample: [ "user::rwx", "group::rwx", "other::rwx" ]
'''

import os
import platform

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native


def split_entry(entry):
    ''' splits entry and ensures normalized return'''

    a = entry.split(':')

    d = None
    if entry.lower().startswith("d"):
        d = True
        a.pop(0)

    if len(a) == 2:
        a.append(None)

    t, e, p = a
    t = t.lower()

    if t.startswith("u"):
        t = "user"
    elif t.startswith("g"):
        t = "group"
    elif t.startswith("m"):
        t = "mask"
    elif t.startswith("o"):
        t = "other"
    else:
        t = None

    return [d, t, e, p]


def build_entry(etype, entity, permissions=None, use_nfsv4_acls=False):
    '''Builds and returns an entry string. Does not include the permissions bit if they are not provided.'''
    if use_nfsv4_acls:
        return ':'.join(['A', 'g' if etype == 'group' else '', entity, permissions + 'tcy'])

    if permissions:
        return etype + ':' + entity + ':' + permissions

    return etype + ':' + entity


def build_command(module, mode, path, follow, default, recursive, recalculate_mask, use_nfsv4_acls, entry=''):
    '''Builds and returns a getfacl/setfacl command.'''
    if mode == 'set':
        cmd = [module.get_bin_path('nfs4_setfacl' if use_nfsv4_acls else 'setfacl', True)]
        cmd.extend(['-a' if use_nfsv4_acls else '-m', entry])
    elif mode == 'rm':
        cmd = [module.get_bin_path('nfs4_setfacl' if use_nfsv4_acls else 'setfacl', True)]
        cmd.extend(['-x', entry])
    else:  # mode == 'get'
        cmd = [module.get_bin_path('getfacl', True)]
        # prevents absolute path warnings and removes headers
        if platform.system().lower() == 'linux':
            if use_nfsv4_acls:
                # use nfs4_getfacl instead of getfacl if use_nfsv4_acls is True
                cmd = [module.get_bin_path('nfs4_getfacl', True)]
            else:
                cmd = [module.get_bin_path('getfacl', True)]
                cmd.append('--absolute-names')
            cmd.append('--omit-header')

    if recursive and not use_nfsv4_acls:
        cmd.append('--recursive')

    if recalculate_mask == 'mask' and mode in ['set', 'rm']:
        cmd.append('--mask')
    elif recalculate_mask == 'no_mask' and mode in ['set', 'rm']:
        cmd.append('--no-mask')

    if not follow and not use_nfsv4_acls:
        if platform.system().lower() == 'linux':
            cmd.append('--physical')
        elif platform.system().lower() == 'freebsd':
            cmd.append('-h')

    if default:
        cmd.insert(1, '-d')

    cmd.append(path)
    return cmd


def acl_changed(module, cmd, entry, use_nfsv4_acls=False):
    '''Returns true if the provided command affects the existing ACLs, false otherwise.'''
    # To check the ACL changes, use the output of setfacl or nfs4_setfacl with '--test'.
    # FreeBSD do not have a --test flag, so by default, it is safer to always say "true".
    if platform.system().lower() == 'freebsd':
        return True

    cmd = cmd[:]  # lists are mutables so cmd would be overwritten without this
    cmd.insert(1, '--test')
    lines = run_acl(module, cmd)
    counter = 0
    for line in lines:
        if not use_nfsv4_acls and not line.endswith('*,*'):
            return True
        # if use_nfsv4_acls and entry is listed
        if use_nfsv4_acls and entry == line:
            counter += 1

    # The current 'nfs4_setfacl --test' lists a new entry,
    # which will be added at the top of the list, followed by the existing entries.
    # So if the entry has already been registered, the entry should be found twice.
    if not use_nfsv4_acls or counter == 2:
        return False
    return True


def run_acl(module, cmd, check_rc=True):
    '''Runs the provided command and returns the output as a list of lines.'''
    try:
        (rc, out, err) = module.run_command(cmd, check_rc=check_rc)
    except Exception as e:
        module.fail_json(msg=to_native(e))

    lines = []
    for l in out.splitlines():
        if not l.startswith('#'):
            lines.append(l.strip())

    if lines and not lines[-1].split():
        # trim last line only when it is empty
        return lines[:-1]

    return lines


def main():
    module = AnsibleModule(
        argument_spec=dict(
            path=dict(type='path', required=True, aliases=['name']),
            entry=dict(type='str'),
            entity=dict(type='str', default=''),
            etype=dict(
                type='str',
                choices=['group', 'mask', 'other', 'user'],
            ),
            permissions=dict(type='str'),
            state=dict(
                type='str',
                default='query',
                choices=['absent', 'present', 'query'],
            ),
            follow=dict(type='bool', default=True),
            default=dict(type='bool', default=False),
            recursive=dict(type='bool', default=False, aliases=['recurse']),
            recalculate_mask=dict(
                type='str',
                default='default',
                choices=['default', 'mask', 'no_mask'],
            ),
            use_nfsv4_acls=dict(type='bool', default=False)
        ),
        supports_check_mode=True,
    )

    if platform.system().lower() not in ['linux', 'freebsd']:
        module.fail_json(msg="The acl module is not available on this system.")

    path = module.params.get('path')
    entry = module.params.get('entry')
    entity = module.params.get('entity')
    etype = module.params.get('etype')
    permissions = module.params.get('permissions')
    state = module.params.get('state')
    follow = module.params.get('follow')
    default = module.params.get('default')
    recursive = module.params.get('recursive')
    recalculate_mask = module.params.get('recalculate_mask')
    use_nfsv4_acls = module.params.get('use_nfsv4_acls')

    if not os.path.exists(path):
        module.fail_json(msg="Path not found or not accessible.")

    if state == 'query':
        if recursive:
            module.fail_json(msg="'recursive' MUST NOT be set when 'state=query'.")

        if recalculate_mask in ['mask', 'no_mask']:
            module.fail_json(msg="'recalculate_mask' MUST NOT be set to 'mask' or 'no_mask' when 'state=query'.")

    if not entry:
        if state == 'absent' and permissions and not use_nfsv4_acls:
            module.fail_json(msg="'permissions' MUST NOT be set when 'state=absent'.")

        if state == 'absent' and not entity:
            module.fail_json(msg="'entity' MUST be set when 'state=absent'.")

        if state in ['present', 'absent'] and not etype:
            module.fail_json(msg="'etype' MUST be set when 'state=%s'." % state)

    if entry:
        if etype or entity or permissions:
            module.fail_json(msg="'entry' MUST NOT be set when 'entity', 'etype' or 'permissions' are set.")

        if state == 'present' and not entry.count(":") in [2, 3]:
            module.fail_json(msg="'entry' MUST have 3 or 4 sections divided by ':' when 'state=present'.")

        if state == 'absent' and not entry.count(":") in [1, 2]:
            module.fail_json(msg="'entry' MUST have 2 or 3 sections divided by ':' when 'state=absent'.")

        if state == 'query':
            module.fail_json(msg="'entry' MUST NOT be set when 'state=query'.")

        default_flag, etype, entity, permissions = split_entry(entry)
        if default_flag is not None:
            default = default_flag

    if platform.system().lower() == 'freebsd':
        if recursive:
            module.fail_json(msg="recursive is not supported on that platform.")

    changed = False
    msg = ""

    if state == 'present':
        entry = build_entry(etype, entity, permissions, use_nfsv4_acls)
        command = build_command(
            module, 'set', path, follow,
            default, recursive, recalculate_mask, use_nfsv4_acls, entry
        )
        changed = acl_changed(module, command, entry, use_nfsv4_acls)

        if changed and not module.check_mode:
            run_acl(module, command)
        msg = "%s is present" % entry

    elif state == 'absent':
        if use_nfsv4_acls:
            entry = build_entry(etype, entity, permissions, use_nfsv4_acls)
        else:
            entry = build_entry(etype, entity, use_nfsv4_acls)
        command = build_command(
            module, 'rm', path, follow,
            default, recursive, recalculate_mask, use_nfsv4_acls, entry
        )
        changed = acl_changed(module, command, entry, use_nfsv4_acls)

        if changed and not module.check_mode:
            run_acl(module, command, False)
        msg = "%s is absent" % entry

    elif state == 'query':
        msg = "current acl"

    acl = run_acl(
        module,
        build_command(
            module, 'get', path, follow, default, recursive,
            recalculate_mask, use_nfsv4_acls
        )
    )

    module.exit_json(changed=changed, msg=msg, acl=acl)


if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function


__metaclass__ = type
import socket

from functools import total_ordering
from itertools import count, groupby


LOGGING_SEVMAP = {
    0: "emergency",
    1: "alert",
    2: "critical",
    3: "error",
    4: "warning",
    5: "notification",
    6: "informational",
    7: "debugging",
}


def search_obj_in_list(name, lst, identifier):
    for o in lst:
        if o[identifier] == name:
            return o
    return None


def flatten_dict(x):
    result = {}
    if not isinstance(x, dict):
        return result

    for key, value in x.items():
        if isinstance(value, dict):
            result.update(flatten_dict(value))
        else:
            result[key] = value

    return result


def validate_ipv4_addr(address):
    address = address.split("/")[0]
    try:
        socket.inet_aton(address)
    except socket.error:
        return False
    return address.count(".") == 3


def validate_ipv6_addr(address):
    address = address.split("/")[0]
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:
        return False
    return True


def normalize_interface(name):
    """Return the normalized interface name"""
    if not name:
        return

    def _get_number(name):
        digits = ""
        for char in name:
            if char.isdigit() or char in "/.":
                digits += char
        return digits

    if name.lower().startswith("et"):
        if_type = "Ethernet"
    elif name.lower().startswith("vl"):
        if_type = "Vlan"
    elif name.lower().startswith("lo"):
        if_type = "loopback"
    elif name.lower().startswith("po"):
        if_type = "port-channel"
    elif name.lower().startswith("nv"):
        if_type = "nve"
    else:
        if_type = None

    number_list = name.split(" ")
    if len(number_list) == 2:
        number = number_list[-1].strip()
    else:
        number = _get_number(name)

    if if_type:
        proper_interface = if_type + number
    else:
        proper_interface = name

    return proper_interface


def get_interface_type(interface):
    """Gets the type of interface"""
    if interface.upper().startswith("ET"):
        return "ethernet"
    elif interface.upper().startswith("VL"):
        return "svi"
    elif interface.upper().startswith("LO"):
        return "loopback"
    elif interface.upper().startswith("MG"):
        return "management"
    elif interface.upper().startswith("MA"):
        return "management"
    elif interface.upper().startswith("PO"):
        return "portchannel"
    elif interface.upper().startswith("NV"):
        return "nve"
    else:
        return "unknown"


def remove_rsvd_interfaces(interfaces):
    """Exclude reserved interfaces from user management"""
    if not interfaces:
        return []
    return [i for i in interfaces if get_interface_type(i["name"]) != "management"]


def vlan_range_to_dict(vlans):
    result = {}
    if vlans:
        for part in vlans.split(","):
            if part == "none":
                break
            if "-" in part:
                a, b = part.split("-")
                a, b = int(a), int(b)
                for vlan in range(a, b + 1):
                    result[str(vlan)] = vlan
            else:
                a = int(part)
                result[str(a)] = a
    return result


def numerical_sort(string_int_list):
    """Sorts list of integers that are digits in numerical order."""

    as_int_list = []

    for vlan in string_int_list:
        as_int_list.append(int(vlan))
    as_int_list.sort()
    return as_int_list


def get_logging_sevmap(invert=False):
    x = LOGGING_SEVMAP
    if invert:
        # cannot use dict comprehension yet
        # since we still test with Python 2.6
        x = dict(map(reversed, x.items()))
    return x


def get_ranges(data):
    """
    Returns a generator object that yields lists of
    consequtive integers from a list of integers.
    """
    for _k, group in groupby(data, lambda t, c=count(): int(t) - next(c)):
        yield list(group)


def vlan_list_to_range(cmd):
    """
    Converts a comma separated list of vlan IDs
    into ranges.
    """
    ranges = []
    for v in get_ranges(cmd):
        ranges.append("-".join(map(str, (v[0], v[-1])[: len(v)])))
    return ",".join(ranges)


def generate_switchport_trunk(type, add, vlans_range):
    """
    Generates a list of switchport commands based on the trunk type and VLANs range.
    Ensures that the length of VLANs lexeme in a command does not exceed 220 characters.
    """

    def append_command():
        command_prefix = f"switchport trunk {type} vlan "
        if add or commands:
            command_prefix += "add "
        commands.append(command_prefix + ",".join(current_chunk))

    commands = []
    current_chunk = []
    current_length = 0

    for vrange in vlans_range.split(","):
        next_addition = vrange if not current_chunk else "," + vrange
        if current_length + len(next_addition) <= 220:
            current_chunk.append(vrange)
            current_length += len(next_addition)
        else:
            append_command()
            current_chunk = [vrange]
            current_length = len(vrange)

    if current_chunk:
        append_command()

    return commands


@total_ordering
class Version:
    """Simple class to compare arbitrary versions"""

    def __init__(self, version_string):
        self.components = version_string.split(".")

    def __eq__(self, other):
        other = _coerce(other)
        if not isinstance(other, Version):
            return NotImplemented

        return self.components == other.components

    def __lt__(self, other):
        other = _coerce(other)
        if not isinstance(other, Version):
            return NotImplemented

        return self.components < other.components


def _coerce(other):
    if isinstance(other, str):
        other = Version(other)
    if isinstance(other, (int, float)):
        other = Version(str(other))
    return other
